#!/usr/bin/env python3

import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console

app = typer.Typer(help="Manage environments.")


def format_environment_data(e, zipped=False):
    create_time = e.create_time.replace(microsecond=0)
    update_time = e.update_time.replace(microsecond=0)

    c = Client()

    environment_id = e.id
    current_project = utils.get_project(c.config.project)
    current_environment = utils.get_environment(current_project, c.config.environment)

    if (
        "@" not in environment_id
        and current_project.endswith(environment_id)
        and (current_environment in ("production") or not current_environment)
    ):
        environment_id = f"{environment_id} (active)"
    elif e.name == utils.get_environment_name(current_project, current_environment):
        environment_id = f"{environment_id} (active)"

    if zipped:
        data = [
            environment_id,
            e.source,
            e.summary.model_count,
            e.summary.model_version_count,
            e.summary.prediction_count,
            create_time,
            update_time,
        ]
        headers = [
            "ID",
            "Source",
            "Models",
            "Model Versions",
            "Predictions",
            "Created",
            "Updated",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            environment_id,
            e.summary.model_count,
            e.summary.model_version_count,
            e.summary.prediction_count,
            create_time,
            # update_time,
        ]
        headers = [
            "ID",
            "Models",
            "Model Versions",
            "Predictions",
            "Created",
            # "Updated",
        ]
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30, project=None, filters=[], style=None)


@app.command("list")
@utils.exit_on_error
def list(
    project: str = typer.Option(None, help="Project ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List environments."""
    project = utils.get_project(project)
    c = Client(project=project)
    projectObj = c.projects.get(c.config.project)
    data = []
    headers = []
    filter_snippet = " (n=%s)" % n
    for e in projectObj.environments.list(n):
        (e_data, headers) = format_environment_data(e)
        data.append(e_data)
    typer.secho(
        "Found %s environments in project %s%s:" % (len(data), project, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    json: bool = typer.Option(False, "--json", help="Print full JSON representation."),
):
    """Get environment details."""
    c = Client(project=project)
    projectObj = c.projects.get(c.config.project)

    env = projectObj.environments.get(environment)
    if json:
        console = Console()
        console.print(env.to_dict())
    else:
        data = format_environment_data(env, zipped=True)
        typer.secho("\nRetrieving environment %s: \n" % (environment), fg="blue")
        utils.print_info(data)


@app.command("create")
@utils.exit_on_error
def create(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option(None, help="Project ID."),
    source: str = typer.Option(
        None,
        "--from",
        help="Project or Environment name from which to create this Environment.",
    ),
):
    """Create environment."""
    c = Client(project=project)
    projectObj = c.projects.get(c.config.project)
    try:
        env = projectObj.environments.create(id=environment, source=source)
        typer.secho("Successfully created environment %s." % env.name, fg="green")
    except Exception as e:
        typer.secho(
            "Failed to create environment %s in project %s: %s"
            % (environment, project, str(e)),
            fg="red",
        )


@app.command("delete")
@utils.exit_on_error
def delete(
    environment: str = typer.Argument(..., help="Environment ID."),
    project: str = typer.Option("", help="Project ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete environment."""
    c = Client(project=project)
    projectObj = c.projects.get(c.config.project)

    if not force:
        force = typer.confirm(
            "Are you sure you want to delete environment %s in project %s? This will delete all resources in the environment."
            % (environment, project)
        )
    if force:
        env = projectObj.environments.get(environment)
        env.delete()
        if c.config.environment == environment and c.config.project == project:
            typer.secho(
                "Since (%s) was the active environment in the active project, the active environment was switched to production."
                % environment,
                fg="blue",
            )
            c.config.set_environment("production")
        typer.secho(
            "Successfully deleted environment %s in project %s."
            % (environment, project),
            fg="green",
        )
    else:
        typer.secho("Environment deletion cancelled! ", fg="green")



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.model_versions import ModelVersionManager
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


from continual.python.sdk.batchpredictions import (
    BatchPredictionManager,
)


class ModelManager(Manager):
    """Manages model resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}"

    def create(
        self,
        display_name: str,
        description: Optional[str] = "",
        get_if_exists: bool = True,
    ) -> Model:
        """Create model.

        Arguments:
            display_name: Dataset name or id.
            description: A brief description of this model.
            get_if_exists: If the model with the display name already exists, get it if this is true, else throw an error.

        Returns
            A Model.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> run.models.create(display_name="my_model", description="Customer churn model")
            <Model object {'name': 'projects/test_proj_4/environments/test_env/models/my_model',
            'description': 'Customer churn model', 'author': 'users/BefwyWcn6x7SNC533zfaAR',
            'display_name': 'my_model', 'create_time': '2022-12-19T16:31:26.638028Z',
            'update_time': '2022-12-19T16:31:26.638028Z', 'state': '', 'current_version': '',
            'latest_model_version': '', 'latest_batch_prediction': ''}>
        """
        req = management_pb2.CreateModelRequest(
            name=self.name(display_name),
            description=description,
            get_if_exists=get_if_exists,
            run=self.run_name,
        )
        resp = self.client._management.CreateModel(req)
        return Model.from_proto(resp, client=self.client, parent_run=self.run_name)

    def get(self, id: str) -> Model:
        """Get model.

        Arguments:
            id: Model name or id.

        Returns
            An experiment.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> run.models.get("my_model")
            <Model object {'name': 'projects/test_proj_4/environments/test_env/models/my_model',
            'description': 'Customer churn model', 'author': 'users/BefwyWcn6x7SNC533zfaAR',
            'display_name': 'my_model', 'create_time': '2022-12-19T16:31:26.638028Z',
            'update_time': '2022-12-19T16:31:26.638028Z', 'state': '', 'current_version': '',
            'latest_model_version': '', 'latest_batch_prediction': ''}>
        """
        req = management_pb2.GetModelRequest(name=self.name(id))
        resp = self.client._management.GetModel(req)
        return Model.from_proto(resp, client=self.client, parent_run=self.run_name)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
        all_projects: bool = False,
    ) -> List[Model]:
        """List model.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
            all_projects: Whether to include all instances of this resource from the project or just from the current parent.

        Returns:
            A list of models.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> models = [run.models.create(display_name=f"my_model_{i}", description="Customer churn model") for i in range(3)]
            >>> [model.display_name for model in run.models.list(page_size=3)]
            ['my_model_2', 'my_model_1', 'my_model_0']
            >>> [model.display_name for model in run.models.list(page_size=3, latest=False)]
            ['my_model_0', 'my_model_1', 'my_model_2']
        """
        req = management_pb2.ListModelsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
            all_projects=all_projects,
        )
        resp = self.client._management.ListModels(req)
        return [
            Model.from_proto(x, client=self.client, parent_run=self.run_name)
            for x in resp.models
        ]

    def list_all(self) -> Pager[Model]:
        """List all model.

        Pages through all model using an iterator.

        Returns:
            A iterator of all model.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> models = [run.models.create(display_name=f"my_model_{i}", description="Customer churn model") for i in range(3)]
            >>> [model.display_name for model in run.models.list_all()]
            ['my_model_0', 'my_model_1', 'my_model_2']
        """

        def next_page(next_page_token):
            req = management_pb2.ListModelsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListModels(req)
            return (
                [
                    Model.from_proto(x, client=self.client, parent_run=self.run_name)
                    for x in resp.models
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class Model(Resource, types.Model):
    """Model resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}"
    _manager: ModelManager
    """Model manager."""

    _model_versions: ModelVersionManager
    """ModelVersion manager."""

    _promotions: PromotionManager
    """Promotion manager."""

    _batch_predictions: BatchPredictionManager
    """Batch Prediction  manager."""

    _tags: TagsManager
    """Tags Manager"""

    _metadata: MetadataManager
    """Metadata Manager"""

    def _init(self):
        self._manager = ModelManager(
            parent=self.parent, client=self.client, run_name=self.parent_run
        )
        self._model_versions = ModelVersionManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._promotions = PromotionManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._batch_predictions = BatchPredictionManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._tags = TagsManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._metadata = MetadataManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )

    @property
    def model_versions(self) -> ModelVersionManager:
        """ModelVersion manager."""
        return self._model_versions

    @property
    def promotions(self) -> PromotionManager:
        """Promotion manager."""
        return self._promotions

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch Prediction manager."""
        return self._batch_predictions

    @property
    def tags(self) -> TagsManager:
        """Tags Manager"""
        return self._tags

    @property
    def metadata(self) -> MetadataManager:
        """Metadata Manager"""
        return self._metadata



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


class PromotionManager(Manager):
    """Manages promotion resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/promotions/{promotion}"

    def create(
        self,
        reason: str,
        model_version_name: str,
    ) -> Promotion:
        """Create promotion.

        Arguments:
            reason: A description of why the model version is being promoted
            model_version_name: The name of the model_version

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> model.promotions.create(model_version_name=model_version.name, reason="UPLIFT")
            <Promotion object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/promotions/ceg9c025lsrt9r5a8l70',
            'create_time': '2022-12-19T16:49:04.118916Z', 'model_version': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg99oq5lsrt9r5a8l2g',
            'reason': 'UPLIFT', 'account': 'users/BefwyWcn6x7SNC533zfaAR',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0', 'base_model_version': '',
            'state': 'STATE_UNSPECIFIED', 'improvement_metric': '', 'improvement_metric_value': 0.0, 'base_improvement_metric_value': 0.0,
            'improvement_metric_diff': 0.0, 'message': ''}>
        """
        req = management_pb2.CreatePromotionRequest(
            model_version_name=model_version_name, run_name=self.run_name, reason=reason
        )
        resp = self.client._management.CreatePromotion(req)
        return Promotion.from_proto(resp, client=self.client)

    def get(self, id: str) -> Promotion:
        """Get promotion.

        Arguments:
            id: Promotion name or id.

        Returns
            A promotion.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> promotion = model.promotions.create(model_version_name=model_version.name, reason="UPLIFT")
            >>> model.promotions.get(promotion.id)
            <Promotion object {'name': 'projects/test_proj_4/environments/test_env/models/my_model/promotions/ceg9c025lsrt9r5a8l70',
            'create_time': '2022-12-19T16:49:04.118916Z', 'model_version': 'projects/test_proj_4/environments/test_env/models/my_model/versions/ceg99oq5lsrt9r5a8l2g',
            'reason': 'UPLIFT', 'account': 'users/BefwyWcn6x7SNC533zfaAR',
            'run_name': 'projects/test_proj_4/environments/test_env/runs/ceg93ji5lsrt9r5a8kt0', 'base_model_version': '',
            'state': 'STATE_UNSPECIFIED', 'improvement_metric': '', 'improvement_metric_value': 0.0, 'base_improvement_metric_value': 0.0,
            'improvement_metric_diff': 0.0, 'message': ''}>
        """
        req = management_pb2.GetPromotionRequest(name=self.name(id))
        resp = self.client._management.GetPromotion(req)
        return Promotion.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
    ) -> List[Promotion]:
        """List promotions.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of Promotions.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> promos = [model.promotions.create(reason="UPLIFT", model_version_name=model_version.name) for _ in range(3)]
            >>> len(model.promotions.list(page_size=10))
            3
        """
        req = management_pb2.ListPromotionsRequest(
            parent=self.parent, page_size=page_size, order_by=order_by, latest=latest
        )
        resp = self.client._management.ListPromotions(req)
        return [Promotion.from_proto(x, client=self.client) for x in resp.promotions]

    def list_all(self) -> Pager[Promotion]:
        """List all promotions.

        Pages through all promotions using an iterator.

        Returns:
            A iterator of all promotions.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> promos = [model.promotions.create(reason="UPLIFT", model_version_name=model_version.name) for _ in range(3)]
            >>> len(model.promotions.list_all())
            3
        """

        def next_page(next_page_token):
            req = management_pb2.ListPromotionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListPromotions(req)
            return (
                [Promotion.from_proto(x, client=self.client) for x in resp.promotions],
                resp.next_page_token,
            )

        return Pager(next_page)


class Promotion(Resource, types.Promotion):
    """Promotion resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/promotions/{promotion}"
    _manager: PromotionManager
    """Promotion Manager."""

    _metadata: MetadataManager
    """Metadata Manager"""

    _tags: TagsManager
    """Tags Manager"""

    def _init(self):
        self._manager = PromotionManager(parent=self.parent, client=self.client)
        self._metadata = MetadataManager(parent=self.name, client=self.client)
        self._tags = TagsManager(parent=self.name, client=self.client)

    @property
    def metadata(self) -> MetadataManager:
        """Metadata Manager"""
        return self._metadata

    @property
    def tags(self) -> TagsManager:
        """Tags Manager"""
        return self._tags



#!/usr/bin/env python3

import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client

from rich.console import Console
from typing import List


app = typer.Typer(help="Manage projects.")


def format_project_data(p, zipped=False):
    create_time = p.create_time.replace(microsecond=0)
    update_time = p.update_time.replace(microsecond=0)
    org = p.organization.split("/")[-1]
    c = Client()
    project_id = p.id
    current_project = c.config.project
    if current_project and current_project == p.name:
        project_id = f"{project_id} (active)"
    data = [
        project_id,
        p.display_name,
        org,
        p.summary.model_count,
        p.summary.model_version_count,
        p.summary.prediction_count,
        create_time,
        # update_time,
    ]
    headers = [
        "ID",
        "Display Name",
        "Organization",
        "Models",
        "Model Versions",
        "Predictions",
        "Created",
        # "Updated",
    ]

    if zipped:
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30, filters=[], style=None)


@app.command("list")
@utils.exit_on_error
def list(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List projects."""
    c = Client()
    data = []
    headers = []
    filter_snippet = " (n=%s)" % n
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    for p in c.projects.list(n, filters=filters):
        (p_data, headers) = format_project_data(p)
        data.append(p_data)
    typer.secho("Found %s projects%s:" % (len(data), filter_snippet), fg="blue")
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    project_id: str = typer.Argument(..., help="Project ID."),
    json: bool = typer.Option(False, "--json", help="Print full JSON representation."),
):
    """Get project details."""
    project_id = utils.get_project(project_id)  # handle passing in project@env
    c = Client(project=project_id)
    project = c.projects.get(project_id)
    if json:
        console = Console()
        console.print(project.to_dict())
    else:
        data = format_project_data(project, zipped=True)
        typer.secho("\nRetrieving project %s: \n" % (project_id), fg="blue")
        utils.print_info(data)


@app.command("create")
@utils.exit_on_error
def create(
    project: str = typer.Argument(..., help="Name of project."),
    org: str = typer.Option(..., help="Name of organization.", show_default=False),
):
    """Create project."""
    try:
        c = Client(verify=False)
        c.config._raise_exception = True
        c.projects.get(project)
        typer.secho(
            "Project %s already exists. Please try another name.\n" % project, fg="red"
        )
    except:
        try:
            proj = utils.get_or_create_project(project, org)
            typer.secho(
                "Successfully created project %s." % (proj.id),
                fg="green",
            )
        except Exception as e:
            typer.secho(
                "Error creating project %s: %s" % (project, str(e)),
                fg="red",
            )
            raise typer.Exit(code=1)


@app.command("delete")
@utils.exit_on_error
def delete(
    project_id: str = typer.Argument(..., help="Project ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete project."""
    project_id = utils.get_project(project_id)
    if not force:
        force = typer.confirm(
            "Are you sure you want to delete the project %s?" % project_id
        )
    if force:
        c = Client(project=project_id)
        project = c.projects.get(project_id)
        project.delete()
        typer.secho("Successfully deleted project %s." % project_id, fg="green")
    else:
        typer.secho("Skipping deletion!")


@app.command("update")
@utils.exit_on_error
def update(
    project_id: str = typer.Argument(..., help="Project ID."),
    new_name: str = typer.Option(None, help="New project name"),
):
    """Update project."""
    project_id = utils.get_project(project_id)
    c = Client(project=project_id)
    project = c.projects.get(project_id)
    if new_name is None:
        new_name = project.display_name
    project.update(display_name=new_name)
    typer.secho("Successfully updated project %s." % project_id, fg="green")



import typer
import logging
import subprocess
import os
import yaml
import pandas as pd
from collections import OrderedDict

from continual.python.sdk.exceptions import BaseException
from continual.python.sdk.client import Client

from enum import Enum
from functools import wraps
from git import Repo

from rich.table import Table
from rich.console import Console
from rich.style import Style
from rich.layout import Layout

logger = logging.getLogger("continual.cli.utils")

continual_style_green = Style(color="rgb(157,209,186)")
continual_style_blue = Style(color="rgb(96,238,252)")
continual_style_purple = Style(color="rgb(63,47,159)")
continual_style_dark_blue = Style(color="rgb(5,27,71)")
continual_style_black = Style(color="rgb(0,0,0)")
continual_style_white = Style(color="rgb(255,255,255)")


class ContinualStyle(Enum):
    WHITE = "WHITE"
    BLACK = "BLACK"
    BLUE = "BLUE"
    GREEN = "GREEN"
    PURPLE = "PURPLE"
    DARK_BLUE = "DARK_BLUE"


class PerformanceMetric(str, Enum):
    RMSE = "rmse"
    MAE = "mae"
    R2 = "r2"
    MAPE = "mape"
    SMAPE = "smape"
    PRECISION = "precision"
    RECALL = "recall"
    ACCURACY = "accuracy"
    ROCAUC = "roc_auc"
    F1 = "f1"
    LOGLOSS = "log_loss"


class SplitType(str, Enum):
    TRAIN = "train"
    VALI = "validation"
    TEST = "test"


def get_style(continual_style):
    if continual_style is None:
        c = Client()
        continual_style = c.config.style
    if not isinstance(continual_style, str):
        continual_style = continual_style.value
    if continual_style.upper() == "WHITE":
        return continual_style_white
    elif continual_style.upper() == "BLACK":
        return continual_style_black
    elif continual_style.upper() == "BLUE":
        return continual_style_blue
    elif continual_style.upper() == "GREEN":
        return continual_style_green
    elif continual_style.upper() == "PURPLE":
        return continual_style_purple
    elif continual_style.upper() == "DARK_BLUE":
        return continual_style_dark_blue


# wrapper function, mainly for convenience.
# but helps prevent calling get_project and get_environment in an undesirable order
def get_project_and_environment(project, environment):
    return (get_project(project), get_environment(project, environment))


# takes project and org display names
# returns project object
# creates project if it doesn't exist
def get_or_create_project(project_name, org_name, project_id=None):
    # get project or create if does not exist
    c = Client(verify=False)
    try:
        if not project_id:
            project_id = "projects/%s" % project_name
            project = get_project_from_display_name(project_name)
        else:
            project = c.projects.get(project_id)
    except Exception:
        org1 = get_or_create_org(org_name)
        project = c.projects.create(
            display_name=project_name,
            organization=org1.name,
        )
    return project


# takes environment display names
# returns environment object
# creates environment if it doesn't exist
def get_or_create_environment(display_name, environment_name=None):
    c = Client(verify=False)
    try:
        if not environment_name:
            env = get_environment_from_display_name(display_name)
        else:
            env = c.projects.get(environment_name)
    except Exception:
        env = c.environments.create(
            id=display_name,
        )
    return env


# returns proj obj of given display_name
# returns first found, if multiple exist w/ same name
# this exists because you can't re-use a project id once assigned,
### so the previous function may change your project id from projects/<project_name>
### to projects/<project_name>-#. This can be unexpected and lead to errors when trying
### to access the project downstream
# essentially, this imposes uniqueness of the display names
def get_project_from_display_name(project_name):
    c = Client(verify=False)
    try:
        projs = c.projects.list()
        proj = [x for x in projs if x.display_name == project_name][0]
        return proj
    except:
        raise


def get_environment_from_display_name(project_name, environment_name):
    c = Client()
    try:
        envs = c.environments.list()
        env = [x for x in envs if x.display_name == environment_name][0]
        return env
    except:
        raise


# takes org display name or optionally id
# returns org object
# creates org if display_name doesn't exist
def get_or_create_org(display_name, id=None):
    c = Client(verify=False)
    try:
        if not id:
            org = get_org_from_display_name(display_name)
        else:
            org = c.organizations.get(id)
    except Exception:
        org = c.organizations.create(display_name)
    return org


# quick fix for dev-464
# returns org id of given display_name
# returns first found, if multiple exist w/ same name
def get_org_from_display_name(org_name):
    c = Client()
    try:
        orgs = c.organizations.list()
        org = [x for x in orgs if x.display_name == org_name][0]
        return org
    except:
        raise


def get_project(project):
    if (project is None) or (project == ""):
        project = get_default_project()
    return project


def get_default_project():
    c = Client()
    proj = c.config.project
    if proj is None or proj == "":
        typer.secho(
            "No project set.  Use continual config set-project or --project to set a project.",
            fg="red",
        )
        exit(1)
    return proj.split("/")[-1]


def get_environment(project, environment):
    if environment is None or environment == "":
        environment = get_default_environment()
    return environment


def get_default_environment():
    c = Client()
    env = c.config.environment
    if env is None or env == "":
        project = c.projects.get(c.config.project)
        env = project.default_environment
    if env is None or env == "":
        env = "production"
    return env


def get_environment_name(project, environment) -> str:
    project = get_project(project)
    environment = get_environment(project, environment)
    if not project.startswith("projects/"):
        project = f"projects/{project}"
    return f"{project}/environments/{environment}"


def get_app_url(client: Client = None):
    if not client:
        client = Client()
    return client.config.app_url


def print_table(
    data,
    headers,
    print=True,
    style=continual_style_green,
    row_styles=["default", "dim"],
    title=None,
    caption=None,
):
    table = Table(
        expand=True, style=style, row_styles=row_styles, title=title, caption=caption
    )
    for x in headers:
        table.add_column(x, overflow="fold")
    if isinstance(data, pd.DataFrame):
        for i, row in data.iterrows():
            table.add_row(*[str(x) for x in row])
    else:
        for row in data:
            table.add_row(*[str(x) for x in row])
    if print:
        console = Console()
        console.print(table)
    return table


def print_info(
    data, show_index=False, headers="", tablefmt="plain", style=continual_style_blue
):
    # typer.echo(tabulate(data, headers=headers, showindex=show_index, tablefmt=tablefmt))
    table = Table(style=style).grid(expand=True)
    for x in headers:
        table.add_column(x, overflow="fold")
    if isinstance(data, pd.DataFrame):
        for i, row in data.iterrows():
            table.add_row(*[str(x) for x in row])
    else:
        for row in data:
            table.add_row(*[str(x) for x in row])
    console = Console()
    console.print(table)


def process_template(template, layout, split_direction):
    children = []
    next_split = {}
    for obj in template:
        data = obj.get("data", None)
        split = obj.get("children", None)
        if data is not None:
            children.append(
                Layout(
                    obj.get("data"),
                    name=obj.get("name"),
                    ratio=obj.get("ratio", 1),
                    minimum_size=5,
                )
            )
        elif split is not None:
            name = obj.get("name")
            direction = obj.get("direction", "horizontal")
            children.append(
                Layout(name=name, ratio=obj.get("ratio", 1), minimum_size=5)
            )
            next_split["%s:%s" % (name, direction)] = split

    if len(children) > 0:
        layout.split(direction=split_direction, *children)

    if len(next_split) > 0:
        for index in next_split:
            parts = index.split(":")
            name = parts[0]
            direction = parts[1]
            process_template(next_split[index], layout[name], direction)


# template is a list of dict. Each dict defines a split or renderable
# renderable: [{"name":"Top", "ratio": 1, data:<renderable>},
# split: {"name":"Bottom", "ratio": 1, "direction": "horizontal", children: [<next_template]}]
#
def build_dashboard(template, direction="horizontal"):
    layout = Layout()
    process_template(template, layout, direction)
    return layout


def execute_cmd(cmd, print_it=True):
    popen = subprocess.Popen(
        [cmd], stdout=subprocess.PIPE, universal_newlines=True, shell=True
    )
    output = popen.stdout.read()
    exit_code = popen.wait()
    if print_it:
        print(output)
        # print("exit code: %s" %exit_code)
    return (output, exit_code)


def split_large_csv(csv, output, max_chunk_size=209715200):
    if not os.path.exists(output):
        os.mkdir(output)
    execute_cmd(
        "split -C %s -d %s %s/%s" % (max_chunk_size, csv, output, csv.split("/")[-1])
    )
    return [file.path for file in os.scandir(output) if file.is_file()]


# for progress bar


def chunker(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def exit_with_error(e):
    """Print error and exit.

    Args:
        e: An exception to print.
    """
    if isinstance(e, BaseException):
        typer.secho("Error: " + str(e.message), fg="red")
        if isinstance(e.details, dict) and len(e.details) > 0:
            typer.secho("Details:", fg="red")
            for key, value in e.details.items():
                typer.secho(f"  {key}: {value}", fg="red")
    else:
        typer.secho("Error: " + str(e), fg="red")
    exit(1)


def exit_on_error(func):
    """Decorator that prints and exits on exceptions.

    Args:
        func: Function to wrap.
    Returns:
        A function that prints error and exit(1) on any exception.
    """

    @wraps(func)
    def function_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.debug("an error occured in command", exc_info=e)
            exit_with_error(e)

    return function_wrapper


# modify yaml presenters to format multi-string queries correctly
class quoted(str):
    pass


def quoted_presenter(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style='"')


yaml.representer.SafeRepresenter.add_representer(quoted, quoted_presenter)


class literal(str):
    pass


def literal_presenter(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


yaml.representer.SafeRepresenter.add_representer(literal, literal_presenter)


def ordered_dict_presenter(dumper, data):
    return dumper.represent_dict(data.items())


yaml.representer.SafeRepresenter.add_representer(OrderedDict, ordered_dict_presenter)


def get_message(message, default):
    if not message:
        message = default
    if not message:
        try:
            repo = Repo(search_parent_directories=True)
            message = repo.git.log(n=1, oneline=True)
            if repo.is_dirty():
                message = message.split(" ")
                message[0] = message[0] + "+dirty"
                message = " ".join(message)
        except Exception:
            pass
    return message



from __future__ import annotations
from typing import List, Optional
from continual.python.sdk.metrics import MetricsManager
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


class ExperimentManager(Manager):
    """Manages experiment resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/versions/{version}/experiments/{experiment}"

    def create(self) -> Experiment:
        """Create experiment.

        Returns
            An experiment.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> model_version.experiments.create()
            <Experiment object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/experiments/ceglm7a5lsrkc0osu120',
            'run_name': 'projects/test_project_1/environments/test_env/runs/test_run', 'state': 'PENDING', 'create_time': '2022-12-20T06:50:05.351093Z',
            'performance_metric_name': '', 'performance_metric_val': 0.0, 'training_config': '', 'error_message': '', 'stack_trace': ''}>
        """
        req = management_pb2.CreateExperimentRequest(
            parent=self.parent, run_name=self.run_name
        )
        resp = self.client._management.CreateExperiment(req)
        return Experiment.from_proto(resp, client=self.client)

    def get(self, id: str) -> Experiment:
        """Get experiment.

        Arguments:
            id: Experiment name or id.

        Returns
            An experiment.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> experiment = model_version.experiments.create()
            >>> model_version.experiments.get(experiment.id)
            <Experiment object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/experiments/ceglm7a5lsrkc0osu120',
            'run_name': 'projects/test_project_1/environments/test_env/runs/test_run', 'state': 'PENDING', 'create_time': '2022-12-20T06:50:05.351093Z',
            'performance_metric_name': '', 'performance_metric_val': 0.0, 'training_config': '', 'error_message': '', 'stack_trace': ''}>
        """
        req = management_pb2.GetExperimentRequest(name=self.name(id))
        resp = self.client._management.GetExperiment(req)
        return Experiment.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
        all_projects: bool = False,
    ) -> List[Experiment]:
        """List experiments.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
            all_projects:  Whether to include all instances of this resource from the project or just from the current parent.

        Returns:
            A list of experiments.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> experiments = [model_version.experiments.create() for _ in range(100)]
            >>> len(model_version.experiments.list(page_size=10))
            10
            >>> len(model_version.experiments.list(page_size=50))
            50
        """
        req = management_pb2.ListExperimentsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
            all_projects=all_projects,
        )
        resp = self.client._management.ListExperiments(req)
        return [Experiment.from_proto(x, client=self.client) for x in resp.experiments]

    def list_all(self) -> Pager[Experiment]:
        """List all experiments.

        Pages through all experiments using an iterator.

        Returns:
            A iterator of all experiments.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> experiments = [model_version.experiments.create() for _ in range(100)]
            >>> len(list(model_version.experiments.list_all()))
            100
        """

        def next_page(next_page_token):
            req = management_pb2.ListExperimentsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListExperiments(req)
            return (
                [
                    Experiment.from_proto(x, client=self.client)
                    for x in resp.experiments
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class Experiment(Resource, types.Experiment):
    """Experiment resource."""

    _manager: ExperimentManager
    """Experiment manager."""

    _metrics: MetricsManager
    """Metrics Manager"""

    _artifacts: ArtifactsManager
    """Artifacts Manager"""

    _tags: TagsManager
    """Tags Manager"""

    _metadata: MetadataManager
    """Metadata Manager"""

    name_pattern: str = "projects/{project}/environments/{environment}/models/{model}/versions/{version}/experiments/{experiment}"

    def _init(self):
        self._manager = ExperimentManager(parent=self.parent, client=self.client)
        self._metrics = MetricsManager(parent=self.name, client=self.client)
        self._artifacts = ArtifactsManager(
            parent=self.name, client=self.client, run_name=self.run_name
        )
        self._tags = TagsManager(parent=self.name, client=self.client)
        self._metadata = MetadataManager(parent=self.name, client=self.client)

    @property
    def metrics(self) -> MetricsManager:
        """Metrics Manager"""
        return self._metrics

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifacts Manager"""
        return self._artifacts

    @property
    def tags(self) -> TagsManager:
        """Tags Manager"""
        return self._tags

    @property
    def metadata(self) -> MetadataManager:
        """Metadata Manager"""
        return self._metadata



from typing import List, Union, Optional
import pandas as pd
from pandas.api.types import (
    is_numeric_dtype,
    is_string_dtype,
    is_bool_dtype,
    is_categorical_dtype,
)
from datetime import datetime
import numpy as np
from collections import Counter
from continual.python.sdk.model_versions import ModelVersion
from sklearn.metrics import confusion_matrix
from sklearn.utils import Bunch

from continual.python.sdk.metadata import Metadata


def _get_non_unique_index_count(df, index_col="ID", time_index_col=None):
    if index_col is not None and index_col in df.columns:
        if time_index_col is not None and time_index_col in df.columns:
            return int(df[[index_col, time_index_col]].duplicated().sum())
        else:
            return int(df[[index_col]].duplicated().sum())
    else:
        return None


def _get_feature_analysis(
    data: pd.DataFrame,
    datetime_columns: List[str],
    index_column: str,
):
    # this function will return count, mean, std, min, Q1, Q2, Q3, and max of all the features in the df
    timestamp_list = []
    numerical_list = []
    string_list = []
    categorical_list = []

    for col in data.columns:
        if datetime_columns is not None and col in datetime_columns:
            null_count = None
            temp_timecol_stats = None
            min_dt_ = None  # Likely to run into "ERROR    NaTType does not support utctimetuple" if these are not separately obtained.
            max_dt_ = None
            if data[col].isna().all():
                temp_timecol = data[col]
                null_count = len(data[col])
                temp_timecol = pd.to_datetime(temp_timecol.fillna("01-01-1970"))
                print("All NaT detected - using epoch start time for data statistics")
                temp_timecol_stats = temp_timecol.describe(
                    datetime_is_numeric=True
                ).to_dict()
                min_dt_ = temp_timecol_stats["min"]
                max_dt_ = temp_timecol_stats["max"]

            data_tcol = pd.to_datetime(data[col], utc=True).dt.tz_localize(None)
            return_dict_ = (
                temp_timecol_stats
                if temp_timecol_stats
                else data_tcol.describe(datetime_is_numeric=True).to_dict()
            )

            return_dict_["column_name"] = str(col)
            return_dict_["distinct_values"] = len(data[col].unique())

            return_dict_["null_count"] = (
                null_count if null_count else int(data_tcol.isnull().sum())
            )

            return_dict_["min"] = min_dt_ if min_dt_ else data_tcol.min()
            return_dict_["max"] = max_dt_ if max_dt_ else data_tcol.max()
            return_dict_.pop("25%")
            return_dict_.pop("50%")
            return_dict_.pop("75%")
            return_dict_.pop("count")
            return_dict_.pop("mean")

            histogram_ = {}
            histogram_["num_undefined"] = return_dict_["null_count"]
            return_histo_ = []

            counts_, bins_ = np.histogram(
                pd.to_datetime(data_tcol.dropna()).astype(int) // 10 ** 9
            )
            bins_ = list(zip(bins_.tolist(), bins_.tolist()[1:]))
            for i, (count_, bin_) in enumerate(zip(counts_, bins_)):
                low_dt = datetime.fromtimestamp(bin_[0])
                high_dt = datetime.fromtimestamp(bin_[1])
                return_histo_.append(
                    {
                        "low_value": low_dt,
                        "high_value": high_dt,
                        "sample_count": int(count_),
                    }
                )
            histogram_["buckets"] = return_histo_
            return_dict_["timestamp_histogram"] = histogram_
            for k, v in return_dict_.items():
                if type(v) == float and np.isnan(v):
                    return_dict_[k] = 0.0
            timestamp_list.append(return_dict_)
            continue

        if col == index_column:
            col_type = _get_column_type(data, col, threshold=0)
        else:
            col_type = _get_column_type(data, col)

        if is_numeric_dtype(col_type):
            temp_col_stats = None
            if data[col].isna().all():
                temp_data = data[col]
                temp_data = temp_data.fillna(0)
                if temp_data.dtype.kind != "f":
                    temp_data = temp_data.astype(float)
                temp_col_stats = temp_data.describe().to_dict()
            return_dict_ = (
                temp_col_stats
                if temp_col_stats
                else pd.to_numeric(data[col]).describe().to_dict()
            )

            return_dict_["column_name"] = str(col)
            return_dict_["percentile25"] = return_dict_.pop("25%")
            return_dict_["percentile50"] = return_dict_.pop("50%")
            return_dict_["percentile75"] = return_dict_.pop("75%")
            return_dict_["count"] = int(return_dict_.pop("count"))
            return_dict_["sd"] = return_dict_.pop("std")
            return_dict_["num_nan"] = int(data[col].isna().sum())
            return_dict_["null_count"] = int(data[col].isnull().sum())
            return_dict_["distinct_values"] = len(data[col].unique())

            # getting histograms
            counts_, bins_ = np.histogram(data[col].dropna().to_list())
            bins_ = list(zip(bins_.tolist(), bins_.tolist()[1:]))

            histogram_ = {}
            histogram_["num_nan"] = return_dict_["num_nan"]
            histogram_["num_undefined"] = return_dict_["null_count"]
            return_histo_ = []
            for i, (count_, bin_) in enumerate(zip(counts_, bins_)):
                return_histo_.append(
                    {
                        "low_value": bin_[0],
                        "high_value": bin_[1],
                        "sample_count": int(count_),
                    }
                )
            histogram_["buckets"] = return_histo_
            return_dict_["num_values_histogram"] = histogram_
            for k, v in return_dict_.items():
                if type(v) == float and np.isnan(v):
                    return_dict_[k] = 0.0
            numerical_list.append(return_dict_)

        elif is_string_dtype(col_type):
            return_dict_ = {
                "column_name": str(col),
                "distinct_values": len(data[col].unique()),  # this is risky!
                "null_count": int(data[col].isna().sum()),
                "min_length": 0,
                "max_length": 0,
            }
            if not data[col].isna().all():
                return_dict_["min_length"] = min(data[col].dropna().str.len())
                return_dict_["max_length"] = max(data[col].dropna().str.len())
            string_list.append(return_dict_)
        elif is_categorical_dtype(col_type) or is_bool_dtype(col_type):
            unique_ = data[col].unique()
            return_dict_ = {
                "column_name": str(col),
                "distinct_values": len(unique_),
                "null_count": data[data[col].isna()].shape[0],
                "categorical_distribution": None,
            }
            counts_ = Counter(data[col].dropna().to_list())
            cd_ = []
            category_count = 0
            histogram_max_size = 30
            remaining_count_name = "Additional Categories"
            remaining_count_sum = 0
            sorted_counts_ = dict(
                sorted(counts_.items(), key=lambda item: item[1], reverse=True)
            )

            for k, v in sorted_counts_.items():
                if category_count > histogram_max_size:
                    remaining_count_sum += v
                else:
                    cd_.append(
                        {
                            "label": str(k),
                            "sample_count": v,
                        }
                    )
                category_count += 1

            if remaining_count_sum > 0:
                cd_.append(
                    {
                        "label": remaining_count_name,
                        "sample_count": remaining_count_sum,
                    }
                )
            if return_dict_["null_count"] > 0:
                cd_.append(
                    {
                        "label": "null",
                        "sample_count": return_dict_["null_count"],
                    }
                )
            return_dict_["categorical_distribution"] = {"categories": cd_}
            categorical_list.append(return_dict_)
    return numerical_list, string_list, categorical_list, timestamp_list


def _get_correlation_matrix(df):
    # this will give a dictionary of dictionaries accessible by res[feat_1][feat_2]
    tmp_df = df.corr().to_dict()
    cm = {
        "display_names": [],
        "rows": [],
    }
    display_names_ = []
    for k, t in tmp_df.items():
        kcol_type = _get_column_type(df, k)
        if not kcol_type:
            continue
        display_names_.append(k)
        t_ = []
        for k2, t2 in t.items():
            col_type = _get_column_type(df, k2)
            if not col_type:
                continue
            # golang json marshalling doesn't support NaN of Inf so replace with 0
            # TODO: may want to create custom golang marshaler to handle.
            if pd.isna(t2):
                t2 = 0
            t_.append(t2)
        cm["rows"].append({"correlations": t_})
    cm["display_names"] = display_names_
    return cm


def _get_column_type(df, col, threshold=150):
    """Returns column type from pandas dataframe"""
    unique_ = df[col].unique()
    if df[col].shape[0] >= threshold:
        if df[col].dtype != "object" and len(unique_) >= threshold:
            return df[col].dtype
        elif df[col].dtype == "object" and len(unique_) >= threshold:
            return pd.StringDtype()
        elif len(unique_) < threshold:
            return pd.CategoricalDtype(categories=unique_, ordered=False)
        else:
            raise NotImplementedError
    else:
        return _get_column_type(df, col, threshold // 2)


def _get_dataset_ranges(
    data: pd.DataFrame,
    columns: List[str],
):
    ranges = []
    for col in columns:
        if not data[col].isna().all():
            data_tcol = pd.to_datetime(data[col]).dt.tz_localize(None)
            start_dt = data_tcol.min()
            end_dt = data_tcol.max()
            ranges.append({"name": col, "start_time": start_dt, "end_time": end_dt})
    return ranges


def create_sklearn_confusion_matrix(
    model_version: ModelVersion,
    display_name: str,
    group_name: str,
    y_true: list,
    y_pred: list,
    labels: list = None,
    sample_weight: list = None,
    normalize: bool = None,
) -> Metadata:
    """Creates confusion matrix metadata on this model version.

    Arguments:
        model_version: The model version to create the confusion matrix on
        display_name: The display name of the confusion matrix
        group_name: A label associated with the group of the confusion matrix
        y_true: Ground truth labels
        y_pred: Predictions from a model version
        labels: Labels
        sample_weight: Optional sample weights
        normalize: Whether to normalize.

    Returns:
        Confusion matrix type metadata

    Examples:
        >>> from continual.python.utils import create_sklearn_confusion_matrix
        >>> ... # Assume continual client is initialized
        >>> run = client.runs.create(description="An example run")
        >>> model_version = run.models.create("example_model").model_versions.create()
        >>> y_true = ["cat", "ant", "cat", "cat", "ant", "bird"]
        >>> y_pred = ["ant", "ant", "cat", "cat", "ant", "cat"]
        >>> labels = ["ant", "bird", "cat"]
        >>> create_sklearn_confusion_matrix(
        ...     model_version=model_version,
        ...     display_name="example_confusion_matrix",
        ...     group_name="test",
        ...     y_true=y_true,
        ...     y_pred=y_pred,
        ...     labels=labels,
        ... )
        <Metadata object {'name': 'projects/delgado_inc/environments/production/models/test_model/versions/cemaiqa5lsrtm2p150j0/metadata/cemaiqa5lsrtm2p150kg',
        'key': 'example_confusion_matrix', 'create_time': '2022-12-28T20:39:05.344420Z', 'update_time': '2022-12-28T20:39:05.344420Z',
        'data': '{"rows": [[2, 0, 0], [0, 0, 1], [1, 0, 2]], "display_name": "example_confusion_matrix"}', 'type': 'CONFUSION_MATRIX', 'group_name': 'test'}>
    """
    matrix = confusion_matrix(
        y_true=y_true,
        y_pred=y_pred,
        labels=labels,
        sample_weight=sample_weight,
        normalize=normalize,
    )
    return model_version.metadata.create(
        key=display_name,
        type="CONFUSION_MATRIX",
        data=dict(display_name=display_name, rows=matrix.astype(np.uint).tolist()),
        group_name=group_name,
    )


def feature_importance_from_sklearn(
    model_version: ModelVersion,
    display_name: str,
    feature_names: List[str],
    importance_result: Union[Bunch, dict],
    dataset_version_names: Optional[List[str]] = None,
) -> Metadata:
    """Creates feature importance metadata from sklearn result.

    Arguments:
        model_version: The model version on which to create the feature importance.
        display_name: The display name of the feature importance.
        feature_names: The names of the features in the order that the importancs are in.
        dataset_version_names: The name of the dataset version on which the feature importance was computed.
        importance_result: The result from sklearn's inspection.permutation_importance method.

    Returns:
        Feature importance type metadata

    Examples:
        >>> from continual.python.utils import feature_importance_from_sklearn
        >>> ... # Assume continual client is initialized
        >>> run = client.runs.create(description="An example run")
        >>> model_version = run.models.create("example_model").model_versions.create()
        >>> dataset_version = run.datasets.create("example_dataset").dataset_versions.create()
        >>> X = [[1, 9, 9], [1, 9, 9], [1, 9, 9], [0, 9, 9], [0, 9, 9], [0, 9, 9]]
        >>> y = [1, 1, 1, 0, 0, 0]
        >>> clf = LogisticRegression().fit(X, y)
        >>> result = permutation_importance(clf, X, y, n_repeats=10, random_state=0)
        >>> feature_importance_from_sklearn(
        ...     model_version=model_version,
        ...     display_name="example_feature_importance",
        ...     feature_names=["feature1", "feature2", "feature3"],
        ...     dataset_version_names=[dataset_version.name],
        ...     importance_result=result,
        ... )
        <Metadata object {'name': 'projects/johnson_mcbride_and_wilk/environments/production/models/test_model/versions/cemakbq5lsrtm2p150sg/metadata/cemakbq5lsrtm2p150u0',
        'key': 'example_feature_importance', 'create_time': '2022-12-28T20:42:23.815825Z', 'update_time': '2022-12-28T20:42:23.815825Z',
        'data': '{"importances": {"feature1": [0.33333333333333337, 0.6666666666666667, 0.33333333333333337, 0.33333333333333337, 0.6666666666666667, 1.0, 0.33333333333333337, 0.33333333333333337, 0.33333333333333337, 0.33333333333333337],
        "feature2": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], "feature3": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]}, "dataset_version_name": "test_dataset_version_name"}', 'type': 'FEATURE_IMPORTANCE', 'group_name': ''}>
    """

    return model_version.metadata.create(
        key=display_name,
        type="FEATURE_IMPORTANCE",
        data={
            "dataset_version_names": dataset_version_names,
            "importances": {
                feature_names[i]: importance_result["importances"].tolist()[i]
                for i in range(len(feature_names))
            },
        },
    )



# Public SDK interface.
from continual.python.sdk.client import Client
from continual.python.sdk import exceptions
from continual.python.sdk.users import User
from continual.python.sdk.organizations import Organization
from continual.python.sdk.projects import Project
from continual.python.sdk.models import Model
from continual.python.sdk.model_versions import ModelVersion
from continual.python.sdk.experiments import Experiment
from continual.python.sdk.environments import Environment
from continual.python.sdk.promotions import Promotion
from continual.python.sdk.runs import Run
from continual.python.sdk.tags import Tag
from continual.python.sdk.metrics import Metric
from continual.python.sdk.metadata import Metadata
from continual.python.sdk.data_profiles import DataProfile
from continual.python.sdk.resource_checks import ResourceCheck
from continual.python.sdk.datasets import Dataset
from continual.python.sdk.dataset_versions import DatasetVersion
from continual.python.sdk.dataset_version_assignments import DatasetVersionAssignment
from continual.python.sdk.artifacts import Artifact


# TODO: Figure out what we want to do with nested types.
from continual.rpc.management.v1.types import (
    ModelVersionState,
    PromotionState,
)

continual = Client(verify=False)
"""Client singleton.

To instantiate a custom client use `continual.Client()`.
"""

__all__ = [
    "continual",
    "Client",
    "exceptions",
    "User",
    "Organization",
    "Project",
    "Connection",
    "Model",
    "ModelVersion",
    "Environment",
    "Experiment",
    "Promotion",
    "ModelVersionState",
    "PromotionState",
    "Run",
    "DataProfile",
    "ResourceCheck",
    "Artifact",
    "Dataset",
    "DatasetVersion",
    "DatasetVersionAssignment",
    "Tag",
    "Metric",
    "Metadata",
]



from __future__ import annotations
from typing import List, Optional

from continual.python.sdk.iterators import Pager
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.rpc.management.v1 import (
    management_pb2,
    types as management_types_py,
)


class TagsManager(Manager):
    """Manages tag resources."""

    # the name pattern for tags depends on the resource it was created for
    name_pattern: str = ""

    def create(
        self,
        key: str,
        value: str,
    ) -> Tag:
        """Create tag.

        Arguments:
            key: the tag key
            value: the tag value

        Returns
            A tag.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> model_version.tags.create(key="test_1", value="test_val")
            <Tag object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/tags/cegm5mq5lsrkc0osu130',
            'key': 'test_1', 'value': 'test_val', 'create_time': '2022-12-20T07:23:07.982232Z', 'update_time': '2022-12-20T07:23:07.982232Z'}>
            >>> model_version.tags.create(key="test_2", value=10)
            <Tag object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/tags/cegm5vq5lsrkc0osu13g',
            'key': 'test_2', 'value': '10', 'create_time': '2022-12-20T07:23:43.273380Z', 'update_time': '2022-12-20T07:23:43.273380Z'}>
            >>> model_version.tags.create(key="test_3", value=0.8)
            <Tag object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/tags/cegm67i5lsrkc0osu140',
            'key': 'test_3', 'value': '0.8', 'create_time': '2022-12-20T07:24:14.532739Z', 'update_time': '2022-12-20T07:24:14.532739Z'}>
        """
        req = management_pb2.CreateTagRequest(
            parent=self.parent,
            key=str(key),
            value=str(value),
        )
        resp = self.client._management.CreateTag(req)
        return Tag.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        latest: bool = True,
        order_by: Optional[str] = None,
    ) -> List[Tag]:
        """List tags.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of tags.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> tags = [model_version.tags.create(key=f"test_{i}", value="test_val") for i in range(3)]
            >>> [t.key for t in model_version.tags.list(page_size=10)]
            ['test_2', 'test_1', 'test_0']
            >>> [t.key for t in model_version.tags.list(page_size=10, latest=False)]
            ['test_0', 'test_1', 'test_2']
            >>> [t.key for t in model_version.tags.list(page_size=10, order_by="key")]
            ['test_0', 'test_1', 'test_2']
        """
        if not self.client:
            print(f"Cannot list tags without client")
            return

        req = management_pb2.ListTagsRequest(
            parent=self.parent,
            page_size=page_size,
            latest=latest,
            order_by=order_by,
        )
        resp = self.client._management.ListTags(req)
        return [Tag.from_proto(x, client=self.client) for x in resp.tags]

    def list_all(self) -> Pager[Tag]:
        """List all tags.

        Pages through all tags using an iterator.

        Returns:
            A iterator of all tag.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> tags = [model_version.tags.create(key=f"test_{i}", value="test_val") for i in range(3)]
            >>> [t.key for t in model_version.tags.list_all()]
            ['test_0', 'test_1', 'test_2']
        """

        def next_page(next_page_token):
            req = management_pb2.ListTagsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListTags(req)
            return (
                [Tag.from_proto(x, client=self.client) for x in resp.tags],
                resp.next_page_token,
            )

        return Pager(next_page)

    def get(self, key: str) -> Tag:
        """Get tag.

        Arguments:
            key: The key associated with the tag

        Return
            A Tag

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> tag = model_version.tags.create(key="test_1", value="test_val")
            >>> model_version.tags.get(key="test_1")
            <Tag object {'name': 'projects/test_project_1/environments/test_env/models/test_model/versions/cegl9qq5lsrkc0osu0ug/tags/cegm5mq5lsrkc0osu130',
            'key': 'test_1', 'value': 'test_val', 'create_time': '2022-12-20T07:23:07.982232Z', 'update_time': '2022-12-20T07:23:07.982232Z'}>
        """

        if not self.client:
            print(f"Cannot fetch tag without client")
            return

        req = management_pb2.GetTagRequest(parent=self.parent, key=key)
        res = self.client._management.GetTag(req)
        return res

    def delete(self, name: str):
        """Delete tag.

        Arguments:
            name: The fully qualified name of the tag

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> run = env.runs.create('run0')
            >>> model_version = run.models.create("example-model").model_versions.create()
            >>> tag = model_version.tags.create(key="test_1", value="test_val")
            >>> len(list(model_version.tags.list_all()))
            1
            >>> model_version.tags.delete(name=tag.name)
            >>> len(list(model_version.tags.list_all()))
            0
        """
        if not self.client:
            print(f"Cannot delete tag without client")
            return

        req = management_pb2.DeleteTagRequest(name=name)
        self.client._management.DeleteTag(req)


class Tag(Resource, management_types_py.Tag):
    """Tag resource."""

    # the name pattern for tags depends on the resource it was created for
    name_pattern: str = ""
    _manager: TagsManager
    """Tags manager."""

    def _init(self):
        self._manager = TagsManager(parent=self.parent, client=self.client)



import os
from pathlib import Path
import yaml
from typing import Optional, List
from omegaconf import OmegaConf
from dataclasses import dataclass


@dataclass
class ContinualConfig:
    endpoint: Optional[str] = None
    email: Optional[str] = None
    api_key: Optional[str] = None
    project: Optional[str] = None
    environment: Optional[str] = None
    style: Optional[str] = None
    raise_exception: Optional[bool] = False
    debug: Optional[bool] = False


class Config:
    """Config for client.

    Reads configuration, in listed order, from:
        * Files:
            * ~/.continual/continual.config
            * ./continual.config
            * ./.continual.config
            * `CONTINUAL_CONFIG` file location from environment variable, if set
        * Environment variables:
            * `CONTINUAL_CONFIG`
            * `CONTINUAL_API_KEY`
            * `CONTINUAL_PROJECT`
            * `CONTINUAL_ENVIRONMENT`
            * `CONTINUAL_ENDPOINT`
            * `CONTINUAL_STYLE`
            * `CONTINUAL_RAISE_EXCEPTION`
            * `CONTINUAL_DEBUG`
        * Config __init__ params (`Config(project=)`)
        * Config setters (`config.project=`), setter methods (`config.set_project()`)
    """

    _endpoint: str = ""

    _email: str = ""

    _api_key: str = ""

    _project: str = ""

    _environment: str = ""

    _style: str = ""

    _raise_exception: bool = False

    _debug: bool = False

    _app_url: str = ""

    _config_file_paths: List[str] = []

    def __init__(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        project: Optional[str] = None,
        environment: Optional[str] = None,
        endpoint: Optional[str] = None,
        style: Optional[str] = None,
        raise_exception: Optional[bool] = False,
        debug: Optional[bool] = False,
    ) -> None:
        """Initialize config.

        Args:
            api_key: API key.
            email: Email address.
            project: Current project.
            environment: Current environment.
            endpoint: Cluster endpoint.
            style: Continual style.
            raise_exception: Raise exception if an error occurs.
            debug: Print stack traces and other debug information
        """

        configs = []
        file_config_paths = []

        default_config = OmegaConf.structured(
            ContinualConfig(
                endpoint="https://sdk.continual.ai",
                email=None,
                api_key=None,
                project=None,
                environment=None,
                style="GREEN",  # cli/utils/ContinualStyle.GREEN
                raise_exception=False,
                debug=False,
            )
        )
        configs.append(default_config)

        home_config_path = os.path.join(Path.home(), ".continual", "continual.yaml")
        if os.path.exists(home_config_path):
            file_config_paths.append(home_config_path)
            try:
                home_config = OmegaConf.load(home_config_path)
                configs.append(home_config)
            except Exception as e:
                print(e)

        working_dir_config_path = os.path.join(os.getcwd(), "continual.yaml")
        if os.path.exists(working_dir_config_path):
            file_config_paths.append(working_dir_config_path)
            try:
                working_dir_config = OmegaConf.load(working_dir_config_path)
                configs.append(working_dir_config)
            except Exception as e:
                print(e)

        working_dir_config_path_2 = os.path.join(os.getcwd(), ".continual.yaml")
        if os.path.exists(working_dir_config_path_2):
            file_config_paths.append(working_dir_config_path_2)
            try:
                working_dir_config_2 = OmegaConf.load(working_dir_config_path_2)
                configs.append(working_dir_config_2)
            except Exception as e:
                print(e)

        if os.environ.get("CONTINUAL_CONFIG"):
            config_file_path = os.environ.get("CONTINUAL_CONFIG")
            if os.path.exists(config_file_path):
                file_config_paths.append(config_file_path)
                try:
                    file_config = OmegaConf.load(config_file_path)
                    configs.append(file_config)
                except Exception as e:
                    print(e)

        OMEGACONF_MISSING = "???"

        env_config = OmegaConf.structured(
            ContinualConfig(
                endpoint=os.environ.get("CONTINUAL_ENDPOINT", OMEGACONF_MISSING),
                email=OMEGACONF_MISSING,
                api_key=os.environ.get("CONTINUAL_API_KEY", OMEGACONF_MISSING),
                project=os.environ.get("CONTINUAL_PROJECT", OMEGACONF_MISSING),
                environment=os.environ.get("CONTINUAL_ENVIRONMENT", OMEGACONF_MISSING),
                style=os.environ.get("CONTINUAL_STYLE", OMEGACONF_MISSING),
                raise_exception=os.environ.get(
                    "CONTINUAL_RAISE_EXCEPTION", OMEGACONF_MISSING
                ),
                debug=os.environ.get("CONTINUAL_DEBUG", OMEGACONF_MISSING),
            )
        )
        configs.append(env_config)

        params_config = OmegaConf.structured(
            ContinualConfig(
                endpoint=endpoint if endpoint else OMEGACONF_MISSING,
                email=email if email else OMEGACONF_MISSING,
                api_key=api_key if api_key else OMEGACONF_MISSING,
                project=project if project else OMEGACONF_MISSING,
                environment=environment if environment else OMEGACONF_MISSING,
                style=style if style else OMEGACONF_MISSING,
                raise_exception=raise_exception
                if raise_exception
                else OMEGACONF_MISSING,
                debug=debug if debug else OMEGACONF_MISSING,
            )
        )
        configs.append(params_config)

        conf = OmegaConf.merge(*configs)

        self._endpoint = conf.endpoint
        self.set_api_key(conf.api_key, False)
        self._email = conf.email
        self.set_project(conf.project, False)
        self.set_environment(conf.environment, False)
        self._style = conf.style
        self._raise_exception = conf.raise_exception
        self._debug = conf.debug
        self._config_file_paths = file_config_paths
        self._app_url = self._get_app_url(self.endpoint)

    @property
    def endpoint(self) -> str:
        """Continual endpoint."""
        return self._endpoint

    @property
    def email(self) -> str:
        """Email address."""
        return self._email

    @property
    def api_key(self) -> str:
        """API key."""
        return self._api_key

    @property
    def project(self) -> str:
        """Default project."""
        return self._project

    @property
    def environment(self) -> str:
        """Default environment."""
        return self._environment

    @property
    def style(self) -> str:
        """Default template style"""
        return self._style

    @property
    def raise_exception(self) -> bool:
        """Raise exception if an error occurs"""
        return self._raise_exception

    @property
    def debug(self) -> bool:
        """Print stack traces and other debug information"""
        return self._debug

    def set_project(self, project: str, save: bool = True) -> str:
        """Set the config project.

        Arguments:
            project: The id or name of the project to use in this config
            save: Whether or not to save this configuration to disk.

        Returns:
            The formatted project name set on config
        """
        if project and "/" not in project:
            project = "projects/" + project

        self._project = project

        if save:
            self.save()

        return self.project

    def set_api_key(self, api_key: str, save: bool = True) -> str:
        """Set the API key.

        Arguments:
            api_key: The API key to use in this config
            save: Whether or not to save this configuration to disk.

        Returns:
            The formatted API key set on config
        """

        self._api_key = api_key

        if save:
            self.save()

        return self.api_key

    def set_environment(self, environment: str, save: bool = True) -> str:
        """Set the environment.

        Arguments:
            environment: The environment to use in this config
            save: Whether or not to save this configuration to disk.

        Returns:
            The formatted environment set on config
        """

        if environment:
            if "/" in environment:
                splits = environment.split("/")
                if len(splits) == 4:
                    self._project = "/".join(splits[:-2])
            self._environment = environment
            if save:
                self.save()
        return self.environment

    def show(self):
        """Print config information."""
        print(f"Config Files:")
        if len(self._config_file_paths) > 0:
            print("\n".join(map(lambda x: f"  {x}", self._config_file_paths)))
        print(f"Email: {self.email or ''}")
        print(f"Endpoint: {self.endpoint or ''}")
        print(f"Project: {self.project or ''}")
        print(f"Environment: {self.environment or ''}")
        print(f"Style: {self.style or ''}")
        print(f"Raise Exception: {self.raise_exception or ''}")
        print(f"Debug: {self.debug or ''}")
        print(f"API Key: {'*' * len(self.api_key) if self.api_key else ''}")

    def save(self):
        """Save config to config file."""

        # TODO: perhaps allow specifying where to save (global/homedir vs local/workingdir)

        if os.environ.get("CONTINUAL_CONFIG"):
            config_file_path = os.environ.get("CONTINUAL_CONFIG")
            if os.path.exists(config_file_path):
                Path(os.path.dirname(config_file_path)).mkdir(
                    parents=True, exist_ok=True
                )
                with open(config_file_path, "w") as outfile:
                    # TODO: consider warning, or not saving, api_key here (since it's a secret)
                    config = dict(
                        email=self.email,
                        api_key=self.api_key,
                        project=self.project,
                        environment=self.environment,
                        endpoint=self.endpoint,
                        style=self.style,
                        raise_exception=self.raise_exception,
                        debug=self.debug,
                    )
                    yaml.dump(config, outfile)
        else:
            with open(os.path.join(os.getcwd(), "continual.yaml"), "w") as outfile:
                config = dict(
                    project=self.project,
                    endpoint=self.endpoint,
                    style=self.style,
                    raise_exception=self.raise_exception,
                    debug=self.debug,
                )
                yaml.dump(config, outfile)
            with open(os.path.join(os.getcwd(), ".continual.yaml"), "w") as outfile:
                config = dict(
                    email=self.email,
                    api_key=self.api_key,
                    environment=self.environment,
                )
                yaml.dump(config, outfile)
                # TODO: consider ensuring that .continual.yaml is git ignored

    def _get_app_url(self, client_endpoint: str):
        app_url = "https://app.continual.ai"
        if (
            "localhost" in client_endpoint.lower()
            or "host.docker.internal" in client_endpoint.lower()
            or "ngrok.io" in client_endpoint.lower()
        ):
            app_url = "http://localhost:8111"
        return app_url



from __future__ import annotations
from typing import List, Optional, Tuple
import os
import io
import requests
import tarfile
import mimetypes
from google.resumable_media import DataCorruption
from google.resumable_media.requests import ResumableUpload

from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.rpc.management.v1 import management_pb2, types

CHUNK_SIZE = 117440512  # 112 MB in bytes, chosen arbitrarily


class ArtifactsManager(Manager):
    """Manages artifact resources."""

    # the name pattern for artifacts depends on the resource it was created for
    name_pattern: str = ""

    def create(
        self,
        id: str = "",
        path: str = "",
        external: bool = False,
        url: str = None,
        upload: bool = True,
        mime_type: str = None,
        update_if_exists: bool = True,
    ) -> Artifact:
        """Create artifact.

        Arguments:
            id: An id to uniquely identify the artifact for a given parent.
            path: Path to the artifact in the local filesystem if not external.
            external: True if this artifact will be stored in a Continual google bucket, false if it is remote.
            url: URL for the artifact. Is a signed URL to a google bucket if artifact is not external,
                 optional user-provided URL if it is an external artifact.
            upload: Whether to upload the artifact specified in `path`. Only valid if external=False.
            mime_type: The mime type of the artifact. If internal artifact, provided value will be
                 overwritten by what is inferred from the file at `path`.
            update_if_exists: If true, update the artifact fields if they already exists.

        Returns:
            An Artifact

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> with open("test.txt", "w") as f:
            ...     f.write("test content")
            >>> dataset_version.artifacts.create(
            ...     id="test-artifact-internal",
            ...     path=os.path.join(os.getcwd(), "test.txt"),
            ...     external=False,
            ...     upload=True
            ... )
            <Artifact object {'name': 'projects/continual_test_proj...", 'external': True ...}>
            >>> dataset_version.artifacts.create(
            ...     id="test-artifact-external",
            ...     external=True,
            ...     url="http://test-url"
            ... )
            <Artifact object {'name': 'projects/continual_test_proj...", 'external': False ...}>
            >>> dataset_version.artifacts.create(
            ...     external=True,
            ...     url="http://test-url"
            ... )
            <Artifact object {'name': 'projects/continual_test_proj.../artifacts/cesceta5lsrkagaitfh0", 'external': False ...}>
        """
        if external or not upload:
            req = management_pb2.CreateArtifactRequest(
                parent=self.parent,
                artifact=Artifact(
                    path=path,
                    external=True,
                    url=url,
                    run=self.run_name,
                    mime_type=mime_type,
                ).to_proto(),
                artifact_id=id,
                update_if_exists=update_if_exists,
            )
            res = self.client._management.CreateArtifact(req)
            return Artifact.from_proto(res, client=self.client)

        elif upload:
            artifact_name = ""
            try:
                file_to_upload = path
                if os.path.isdir(path):
                    tarfile_name = os.path.basename(path) + ".tar.gz"
                    with tarfile.open(tarfile_name, "w:gz") as tar:
                        tar.add(
                            path,
                            recursive=True,
                            arcname=os.path.basename(tarfile_name).split(".")[0],
                        )
                    file_to_upload = tarfile_name

                mime_type, _ = mimetypes.guess_type(file_to_upload)

                payload = ""
                with open(file_to_upload, "rb") as f:
                    payload = f.read()

                total_bytes = len(payload)
                exceeds_chunk_size = total_bytes >= CHUNK_SIZE

                req = management_pb2.GenerateArtifactUploadURLRequest(
                    parent=self.parent,
                    artifact_id=id,
                    path=path,
                    mime_type=mime_type or "",
                    resumable=exceeds_chunk_size,
                    run=self.run_name,
                    update_if_exists=update_if_exists,
                )
                res = self.client._management.GenerateArtifactUploadURL(req)

                upload_url = res.url
                artifact_name = res.artifact.name

                headers = dict()
                if mime_type:
                    headers["Content-Type"] = mime_type

                if exceeds_chunk_size:
                    transport = requests.Session()

                    stream = io.BytesIO(payload)

                    upload = ResumableUpload(upload_url, CHUNK_SIZE)
                    upload._resumable_url = upload_url
                    upload._total_bytes = total_bytes
                    upload._stream = stream

                    while not upload.finished:
                        try:
                            print("Writing chunk ... ")
                            response = upload.transmit_next_chunk(transport, timeout=60)
                            print(f"Chunk response: {response}")
                        except DataCorruption:
                            raise
                else:
                    upload_res = requests.put(
                        upload_url, data=payload, headers={"Content-Type": mime_type}
                    )
                    upload_res.raise_for_status()

                return Artifact.from_proto(res.artifact, client=self.client)
            except:
                if artifact_name:
                    req = management_pb2.DeleteArtifactRequest(name=artifact_name)
                    res = self.client._management.DeleteArtifact(req)
                raise

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[Artifact]:
        """List artifacts.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list of artifacts.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of artifacts.

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> artifacts = [dataset_version.artifacts.create(key=f"test_artifact_{i}", path="example_path/test.txt") for i in range(5)]
            >>> [a.key for a in dataset_version.artifacts.list(page_size=2)]
            ['test_artifact_4', 'test_artifact_3']
            >>> [a.key for a in dataset_version.artifacts.list(page_size=2, latest=False)]
            ['test_artifact_0', 'test_artifact_1']
        """
        req = management_pb2.ListArtifactsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListArtifacts(req)
        return [Artifact.from_proto(x, client=self.client) for x in resp.artifacts]

    def get(self, id: str = "") -> Artifact:
        """Get artifact.

        Arguments:
            id: A unique id for the artifact.

        Returns:
            An artifact.

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> artifacts = [dataset_version.artifacts.create(id=f"test-artifact-{i}", path="example_path/test.txt") for i in range(5)]
            >>> dataset_version.artifacts.get(id="test-artifact-0")
            <Artifact object {'name': 'projects/continual_test_proj...", 'id': 'test-artifact-0', 'external': False ...}>
        """
        if not self.client:
            print(f"Cannot fetch artifact without client")
            return

        req = management_pb2.GetArtifactRequest(
            name=self.name(id, self.parent, f"{self.parent}/artifacts/{id}")
        )
        res = self.client._management.GetArtifact(req)
        return Artifact.from_proto(res, client=self.client)

    def delete(self, id: str):
        """Delete artifact.

        Arguments:
            id: The id or fully qualified name of the artifact

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> artifacts = [dataset_version.artifacts.create(key=f"test-artifact-{i}", path="example_path/test.txt") for i in range(5)]
            >>> [dataset_version.artifacts.delete(a.name) for a in dataset_version.artifacts.list(page_size=5)]
            [None, None, None, None, None]
            >>> len(dataset_version.artifacts.list(page_size=5))
            0
        """
        if not self.client:
            print(f"Cannot delete artifact without client")
            return

        req = management_pb2.DeleteArtifactRequest(
            name=self.name(id, self.parent, f"{self.parent}/artifacts/{id}")
        )
        self.client._management.DeleteArtifact(req)

    def download(
        self,
        id: str,
        download_dir: str = "./artifacts",
    ) -> Tuple[Artifact, str]:
        """Download artifact.

        Arguments:
            id: The name or id of the artifact
            download_dir: The directory to which to download the artifact

        Returns:
            A tuple of the Artifact object and the path where it was downloaded.

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> with open("test.txt", "w") as f:
            ...     f.write("test content")
            >>> artifact = dataset_version.artifacts.create(
            ...     key="test-artifact-internal",
            ...     path=os.path.join(os.getcwd(), "test.txt"),
            ...     external=False,
            ...     type="txt",
            ...     metadata={"key": "value"},
            ...     upload=True
            ... )
            >>> artifact, dest_path = dataset_version.artifacts.download(id=artifact.id) # Download the artifact
            >>> with open(dest_path, "r") as f:
            ...     print(f.read())
            "test content"
        """

        artifact = self.get(id)
        downloaded_to = ""
        if not artifact.url:
            raise ValueError(
                f"Artifact cannot be downloaded - no URL was found: {artifact.url}"
            )
        elif artifact.external:
            raise ValueError(f"Cannot download an external artifact - {artifact.url}")

        try:
            res = requests.get(artifact.url, stream=True)
            with tarfile.open(fileobj=res.raw, mode="r") as f:
                f.extractall(download_dir)

            root_dir = os.path.basename(artifact.path or "").split(".")[0]
            downloaded_to = os.path.join(download_dir, root_dir)
        except:
            res = requests.get(artifact.url)
            downloaded_to = os.path.join(
                download_dir,
                os.path.basename(artifact.path or artifact.name.split("/")[-1]),
            )
            with open(downloaded_to, "wb") as f:
                f.write(res.content)
        return artifact, downloaded_to


class Artifact(Resource, types.Artifact):
    """Artifact resource."""

    # the name pattern for artifacts depends on the resource it was created for
    name_pattern: str = ""

    _manager: ArtifactsManager
    """Artifact manager."""

    def _init(self):
        self._manager = ArtifactsManager(parent=self.parent, client=self.client)

    def download(self, dest_dir: str = "./artifacts") -> Tuple[Artifact, str]:
        """Download artifact.

        Arguments:
            dest_dir: The directory to which to download the artifact

        Returns:
            A tuple of the Artifact object and the path where it was downloaded.

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> with open("test.txt", "w") as f:
            ...     f.write("test content")
            >>> artifact = dataset_version.artifacts.create(
            ...     key="test-artifact-internal",
            ...     path=os.path.join(os.getcwd(), "test.txt"),
            ...     external=False,
            ...     type="txt",
            ...     metadata={"key": "value"},
            ...     upload=True
            ... )
            >>> artifact, dest_path = artifact.download() # Download the artifact
            >>> with open(dest_path, "r") as f:
            ...     print(f.read())
            "test content"
        """
        # Create if the default doesnt exist
        if dest_dir == "./artifacts":
            dest_dir = os.path.join(os.getcwd(), "artifacts")
            os.makedirs(dest_dir, exist_ok=True)
        return self._manager.download(id=self.name, download_dir=dest_dir)

    def delete(self):
        """Delete artifact.

        Examples:
            >>> ... # Assume we have an environment defined
            >>> run = env.runs.create("my-run")
            >>> dataset = run.datasets.create("test_dataset")
            >>> dataset_version = dataset.dataset_versions.create()
            >>> artifacts = [dataset_version.artifacts.create(key=f"test_artifact_{i}", path="example_path/test.txt") for i in range(5)]
            >>> [a.delete() for a in dataset_version.artifacts.list(page_size=5)]
            [None, None, None, None, None]
            >>> len(dataset_version.artifacts.list(page_size=5))
            0
        """
        self._manager.delete(name=self.name)



from __future__ import annotations
from typing import Optional
import types
import traceback
from functools import wraps


def handle_error(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        try:
            x = func(self, *args, **kwargs)
        except Exception as e:
            local_client = self.client
            errors = []
            if local_client:
                # TODO: add more specific checks and error messages
                if (
                    local_client.config.api_key is None
                    or local_client.config.api_key == ""
                ):
                    errors.append("No API key set")
                if (
                    local_client.config.project is None
                    or local_client.config.project == ""
                ):
                    errors.append("No project set")
                if local_client.config.debug == True:
                    traceback.print_exc()

                errors.append(str(e))
                if local_client.config.raise_exception:
                    raise Exception(f"{'. '.join(errors)}")
                else:
                    print(f"Errors: {'. '.join(errors)}")
            else:
                raise Exception(f"No client found on {self}")
            return
        return x

    return wrapped


class ManagerMeta(type):
    def __new__(cls, name, bases, namespace, **kwds):
        namespace = {
            k: v
            if (k.startswith("__") or not isinstance(v, types.FunctionType))
            else handle_error(v)
            for k, v in namespace.items()
        }
        return type.__new__(cls, name, bases, namespace)


class Manager(metaclass=ManagerMeta):
    """Base class for resource managers"""

    name_pattern: str
    parent: str
    client: client.Client

    def __init__(
        self, client: client.Client, parent: str = "", run_name: str = ""
    ) -> None:
        self.client = client
        self.parent = parent
        self.run_name = run_name

    def name(
        self, id: str, parent: Optional[str] = None, name_pattern: Optional[str] = None
    ) -> str:
        """Generates a resource name from parent and id.

        Adds wildcard for any missing parent elements.

        Arguments:
            id: Name or id of resource.
            parent: (Optional) Override parent name.
            name_pattern: (Optional) Override name pattern.
        """
        if "/" in id:
            # Don't allow names to override manager parent config since this is confusing
            # and is typically a bug in the user code.
            if parent is not None and parent != "" and not id.startswith(parent):
                raise ValueError(f"Resource {id} not a child of {parent}.")
            return id

        name_pattern = name_pattern or self.name_pattern or ""
        pattern_parts = name_pattern.split("/")
        parent = parent or self.parent or ""
        parent_parts = parent.split("/")
        out = []
        for i in range(len(pattern_parts)):
            if (i + 1) % 2 == 0:
                if i < len(parent_parts):
                    out.append(parent_parts[i])
                else:
                    out.append("-")
            else:
                out.append(pattern_parts[i])
        out[-1] = id
        return "/".join(out)



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from google.protobuf import field_mask_pb2


class UserManager(Manager):
    """Manages user resources."""

    name_pattern: str = "users/{user}"

    def get(self, id: str) -> User:
        """Get user.

        Arguments:
            id: Fully qualified user name or id.

        Returns
            A User.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> client.users.get(id='zxNbTXkbxLeyjb3SUhJ2fR')
            User object {'name': 'users/zxNbTXkbxLeyjb3SUhJ2fR', 'email': 'test@continual.ai',
            'email_verified': True, 'full_name': 'test user', 'update_time': '2022-12-15T01:00:20.707583Z',
            'create_time': '2022-12-15T01:00:20.707583Z', 'trial_available': True, 'first_name':
            'test', 'last_name': 'user', 'bio': '', 'location': '', 'password': '',
            'service_account': False, 'disabled': False}>
        """
        req = management_pb2.GetUserRequest(name=self.name(id))
        resp = self.client._management.GetUser(req)
        return User.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
    ) -> List[User]:
        """List users.

        Arguments:
            page_size: Number of itmes to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of users.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config (first user)
            >>> second_user = client.register(first_name='second', last_name='user', email='test2@continual.ai', password='test123')
            >>> [u.email for u in client.users.list(page_size=10)] # Orders by 'create_time' by default
            ['test@continual.ai', 'test2@continual.ai']
            >>> [u.email for u in client.users.list(page_size=10, latest=False)] # Ascending order
            ['test2@continual.ai', 'test@continual.ai']
        """
        req = management_pb2.ListUsersRequest(
            page_size=page_size, order_by=order_by, latest=latest
        )
        resp = self.client._management.ListUsers(req)
        return [User.from_proto(u, client=self.client) for u in resp.users]

    def list_all(self) -> Pager[User]:
        """List all users.

        Pages through all users using an iterator.

        Returns:
            A iterator of all users.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config (first user)
            >>> client.register(first_name='second', last_name='user', email='test2@continual.ai', password='test123') # Second user
            >>> client.users.list_all()
            <continual.python.sdk.iterators.Pager object at 0x7f82d469f8b0>
            >>> [u.email for u in client.users.list_all()]
            ['test2@continual.ai', 'test@continual.ai']
        """

        def next_page(next_page_token):
            req = management_pb2.ListUsersRequest(page_token=next_page_token)
            resp = self.client._management.ListUsers(req)
            return (
                [User.from_proto(u, client=self.client) for u in resp.users],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete a user. (Admin only)

        Arguments:
            id: Name or id.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config (first user)
            >>> second_user = client.register(first_name='second', last_name='user', email='test2@continual.ai', password='test123') # Second user
            >>> client.users.delete(id=second_user.name)
        """

        req = management_pb2.DeleteUserRequest(name=self.name(id))
        self.client._management.DeleteUser(req)

    def update(
        self,
        id: str,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        bio: Optional[str] = None,
        location: Optional[str] = None,
    ) -> User:
        """Update user.

        Arguments:
            id: The user id
            first_name:  First name of display name.
            last_name:  Last name of display name.
            bio: Bio.
            location: Location.

        Returns:
            Updated user.
        """
        paths = []
        if first_name is not None:
            paths.append("first_name")
        if last_name is not None:
            paths.append("last_name")
        if bio is not None:
            paths.append("bio")
        if location is not None:
            paths.append("location")
        req = management_pb2.UpdateUserRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            user=User(
                name=self.name(id),
                first_name=first_name,
                last_name=last_name,
                bio=bio,
                location=location,
            ).to_proto(),
        )
        resp = self.client._management.UpdateUser(req)
        return User.from_proto(resp, client=self.client)


class User(Resource, types.User):
    """User resource."""

    name_pattern = "users/{users}"

    _manager: UserManager
    """Users Manager."""

    def _init(self):
        self._manager = UserManager(parent=self.parent, client=self.client)

    def delete(self) -> None:
        """Delete user."""
        self._manager.delete(self.name)

    def update(
        self,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        bio: Optional[str] = None,
        location: Optional[str] = None,
    ) -> User:
        """Update user.

        Arguments:
            first_name:  First name of display name.
            last_name:  Last name of display name.
            bio: Bio.
            location: Location.

        Returns:
            Updated user.

        Examples:
            >>> from continual import Client
            >>> client = Client(verify=False)
            >>> user = client.register(first_name='test', last_name='user', email='test@continual.ai', password='test123')
            >>> user.first_name, user.last_name
            ('test', 'user')
            >>> updated_user = user.update(first_name='not test')
            >>> updated_user.first_name, updated_user.last_name
            ('not_test', 'user')
        """
        return self._manager.update(
            self.name,
            first_name=first_name,
            last_name=last_name,
            bio=bio,
            location=location,
        )



from __future__ import annotations
from typing import List, Optional

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager


class DatasetVersionAssignmentManager(Manager):
    """Manages DatasetVersionAssignment resources."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/versions/{version}/assignments/{assignment}"

    def create(self, resource_name: str) -> DatasetVersionAssignment:
        """Create a dataset version assignment.

        A dataset version assignment associates a dataset version with a resource such as a model version or batch prediction.

        Argument:
            resource_name: The resource that is downstream of this parent dataset version

        Returns
            A DatasetVersionAssignment.

        Examples:
            >>> ... # Assume dataset_version and model_version are defined
            >>> dataset_version.assignments.create(resource_name=model_version.name)
            <DatasetVersionAssignment object {'name': 'projects/continual_test_proj/environments/production/datasets/test_dataset/versions/cegig6q5lsrt9r5a8nl0/assignments/cegigs25lsrt9r5a8no0',
            'resource_name': 'projects/continual_test_proj/environments/production/models/test_model/versions/cegig6q5lsrt9r5a8nmg', 'create_time': '2022-12-20T03:13:52.354991Z'}>
        """
        req = management_pb2.CreateDatasetVersionAssignmentRequest(
            dataset_version_name=self.parent,
            resource_name=resource_name,
        )
        resp = self.client._management.CreateDatasetVersionAssignment(req)
        return DatasetVersionAssignment.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[DatasetVersionAssignment]:
        """List dataset version assignments.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of DatasetVersionAssignments.

        Examples:
            >>> ... # Assume run is defined
            >>> dataset_version = run.datasets.create("test_dataset").dataset_versions.create()
            >>> model_versions = [
                    run.models.create(f"test_model_{i}").model_versions.create()
                    for i in range(5)
                ]
            >>> dv_assignments = [
                    dataset_version.assignments.create(resource_name=mv.name)
                    for mv in model_versions
                ]
            >>> len(dataset_version.assignments.list(page_size=10))
            5
        """
        req = management_pb2.ListDatasetVersionAssignmentsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListDatasetVersionAssignments(req)
        return [
            DatasetVersionAssignment.from_proto(x, client=self.client)
            for x in resp.assignments
        ]

    def list_all(self) -> Pager[DatasetVersionAssignment]:
        """List all dataset version assignments.

        Pages through all dataset versions using an iterator.

        Returns:
            A iterator of all DatasetVersionAssignment.

        Examples:
            >>> ... # Assume run is defined
            >>> dataset_version = run.datasets.create("test_dataset").dataset_versions.create()
            >>> model_versions = [
                    run.models.create(f"test_model_{i}").model_versions.create()
                    for i in range(5)
                ]
            >>> dv_assignments = [
                    dataset_version.assignments.create(resource_name=mv.name)
                    for mv in model_versions
                ]
            >>> len(list(dataset_version.assignments.list_all()))
            5
        """

        def next_page(next_page_token):
            req = management_pb2.ListDatasetVersionAssignmentsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListDatasetVersionAssignments(req)
            return (
                [
                    DatasetVersionAssignment.from_proto(x, client=self.client)
                    for x in resp.assignments
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete a dataset version assignment.

        Arguments:
            id: DatasetVersionAssignment name or id.

        Examples:
            >>> ... # Assume dataset_version and model_version are defined
            >>> dv_assignment = dataset_version.assignments.create(resource_name=model_version.name)
            >>> len(dataset_version.assignments.list_all())
            1
            >>> dataset_version.assignments.delete(id=dv_assignment.id)
            >>> len(list(dataset_version.assignments.list_all()))
            0
        """

        req = management_pb2.DeleteDatasetVersionAssignmentRequest(name=self.name(id))
        self.client._management.DeleteDatasetVersionAssignment(req)


class DatasetVersionAssignment(Resource, types.DatasetVersionAssignment):
    """Dataset version resource."""

    name_pattern: str = "projects/{project}/environments/{environment}/datasets/{dataset}/versions/{version}/assignments/{assignment}"

    _manager: DatasetVersionAssignmentManager
    """DatasetVersionAssignment manager."""

    def _init(self):
        self._manager = DatasetVersionAssignmentManager(
            parent=self.parent, client=self.client
        )



from __future__ import annotations
from typing import List, Optional, Any, Union
from datetime import datetime

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager


class MetricsManager(Manager):
    """Manages metric resources."""

    # the name pattern for metrics depends on the resource it was created for
    name_pattern: str = ""

    def create(
        self,
        id: str,
        display_name: str = "",
        direction: str = "UNSPECIFIED",
        update_if_exists: bool = True,
    ) -> Metric:
        """Create a Metric.

        Arguments:
            id: identifier for the metric (e.g. "accuracy")
            display_name: the display name of the metric
            direction: direction of the metric (e.g. "HIGHER", "LOWER", or "UNSPECIFIED")
            update_if_exists: if the metric exists and this is true, update the current metric and return, else raise an error if the metric already exists

        Returns:
            A new metric.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> mape_metric = model_version.metrics.create(id="accuracy", display_name="my accuracy")
            >>> mape_metric.log(value=0.1, group_name="test")
            >>> model_version.metrics.get("accuracy")
            <Metric object {'name': 'projects/test_project/environments/my_environment/models/my_model/versions/cerkggdvn3d3cn5absgg/metrics/accuracy', 'display_name': 'my accuracy', 'create_time': '2023-01-05T21:59:30.214316Z', 'update_time': '2023-01-05T21:59:30.214316Z', 'values': [{'name': 'projects/test_project/environments/production/models/test_model/versions/cerkggdvn3d3cn5absgg/metrics/mape/metricValues/cerkgglvn3d3cn5abss0', 'value': 0.1, 'timestamp': '0001-01-01T00:00:00Z', 'create_time': '2023-01-05T21:59:30.223377Z', 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'step': '0', 'group': ''}], 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'direction': 'UNSPECIFIED'}>
        """

        req = management_pb2.CreateMetricRequest(
            metric=types.Metric(
                display_name=display_name,
                direction=direction,
                run=self.run_name,
            ).to_proto(),
            metric_id=id,
            parent=self.parent,
            update_if_exists=update_if_exists,
        )

        resp = self.client._management.CreateMetric(req)
        return Metric.from_proto(resp, client=self.client)

    def log(
        self,
        metric_name: str,
        value: Any,
        group: str = "",
        timestamp: Union[str, datetime] = None,
        step: int = 0,
        label: str = "",
    ):
        """Log a Metric Value.

        Arguments:
            metric_name: the name of the metric to log this value for
            value: value of the metric
            group: [Optional] name of the group to which this metric belongs (e.g. "train")
            timestamp: [Optional] timestamp for which the value has been logged for
            step: [Optional] step at which the metric was logged, allows metrics to be grouped into a sequence
            label: [Optional] a label that describes the value
        Returns:

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> mape_metric = model_version.metrics.create(id="accuracy", display_name="my accuracy")
            >>> mape_metric.log(value=0.1, group_name="test")
            >>> model_version.metrics.get("accuracy")
            <Metric object {'name': 'projects/test_project/environments/my_environment/models/my_model/versions/cerkggdvn3d3cn5absgg/metrics/accuracy', 'display_name': 'my accuracy', 'create_time': '2023-01-05T21:59:30.214316Z', 'update_time': '2023-01-05T21:59:30.214316Z', 'values': [{'name': 'projects/test_project/environments/production/models/test_model/versions/cerkggdvn3d3cn5absgg/metrics/mape/metricValues/cerkgglvn3d3cn5abss0', 'value': 0.1, 'timestamp': '0001-01-01T00:00:00Z', 'create_time': '2023-01-05T21:59:30.223377Z', 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'step': '0', 'group': ''}], 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'direction': 'UNSPECIFIED'}>
        """
        if isinstance(timestamp, str):
            timestamp = datetime.fromisoformat(timestamp)
        req = management_pb2.LogMetricRequest(
            parent=metric_name,
            value=types.MetricValue(
                value=float(value) if isinstance(value, int) else value,
                group=group,
                step=step,
                timestamp=timestamp,
                run=self.run_name,
                label=label,
            ).to_proto(),
        )
        self.client._management.LogMetric(req)
        return

    def get(self, id: str) -> Metric:
        """Get metric.

        Arguments:
            name: Metric name or id.
            key: Metric key.
            group_name: Metric group name.
            step: Metric step for sequences.

        Returns
            A Metric.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> metric_keys = ['accuracy', 'precision', 'recall']
            >>> metrics = [model_version.metrics.create(key=key, value=0.9, direction="HIGHER", group_name="train") for key in metric_keys]
            >>> model_version.metrics.get(name=metrics[1].name).key                     # get by name
            'precision'
            >>> model_version.metrics.get(key="accuracy", group_name="train").key       # get by key and group_name
            'accuracy'
        """
        req = management_pb2.GetMetricRequest(
            name=self.name(
                id=id,
                parent=self.parent,
                name_pattern=self.parent + "/metrics/{metric}",
            ),
        )
        metric = self.client._management.GetMetric(req)
        return Metric.from_proto(metric, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[Metric]:
        """List metrics.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of metrics.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> metric_keys = ['accuracy', 'precision', 'recall']
            >>> metrics = [model_version.metrics.create(key=key, value=0.9, direction="HIGHER", group_name="train") for key in metric_keys]
            >>> [m.key for m in model_version.metrics.list(page_size=3)]
            ['recall', 'precision', 'accuracy']
            >>> [m.key for m in model_version.metrics.list(page_size=3, order_by="key", latest=False)]
            ['accuracy', 'precision', 'recall']
        """
        req = management_pb2.ListMetricsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListMetrics(req)
        return [Metric.from_proto(x, client=self.client) for x in resp.metrics]

    def list_all(self) -> Pager[Metric]:
        """List all metrics.

        Pages through all metrics using an iterator.

        Returns:
            A iterator of all metrics.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> metric_keys = ['accuracy', 'precision', 'recall']
            >>> metrics = [model_version.metrics.create(key=key, value=0.9, direction="HIGHER", group_name="train") for key in metric_keys]
            >>> [m.key for m in model_version.metrics.list_all()]
            ['accuracy', 'precision', 'recall']
        """

        def next_page(next_page_token):
            req = management_pb2.ListMetricsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListMetrics(req)
            return (
                [Metric.from_proto(x, client=self.client) for x in resp.metrics],
                resp.next_page_token,
            )

        return Pager(next_page)


class Metric(Resource, types.Metric):
    """Metric resource."""

    # the name pattern for metrics depends on the resource it was created for
    name_pattern: str = ""

    _manager: MetricsManager
    """Metrics manager."""

    def _init(self):
        self._manager = MetricsManager(
            parent=self.parent, client=self.client, run_name=self.run
        )

    def log(
        self,
        value: Any,
        group: str = "",
        timestamp: Union[str, datetime] = None,
        step: int = 0,
        label: str = "",
    ):
        """Log a Metric Value for the current Metric.

        Arguments:
            value: value of the metric
            group: [Optional] name of the group to which this metric belongs (e.g. "train")
            timestamp: [Optional] timestamp for which the value has been logged for
            step: [Optional] step at which the metric was logged, allows metrics to be grouped into a sequence
            label: [Optional] a label that describes the value

        Returns:

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> model = run.models.create(display_name="my_model", description="Customer churn model")
            >>> model_version = model.model_versions.create()
            >>> mape_metric = model_version.metrics.create(id="accuracy", display_name="my accuracy")
            >>> mape_metric.log(value=0.1, group_name="test")
            >>> model_version.metrics.get("accuracy")
            <Metric object {'name': 'projects/test_project/environments/production/models/test_model/versions/cerkggdvn3d3cn5absgg/metrics/accuracy', 'display_name': 'my accuracy', 'create_time': '2023-01-05T21:59:30.214316Z', 'update_time': '2023-01-05T21:59:30.214316Z', 'values': [{'name': 'projects/test_project/environments/production/models/test_model/versions/cerkggdvn3d3cn5absgg/metrics/mape/metricValues/cerkgglvn3d3cn5abss0', 'value': 0.1, 'timestamp': '0001-01-01T00:00:00Z', 'create_time': '2023-01-05T21:59:30.223377Z', 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'step': '0', 'group': ''}], 'created_by': 'projects/test_project/apikeys/ceeeg5lvn3d2lroffjv0', 'run': '', 'direction': 'UNSPECIFIED'}>
        """
        self._manager.log(
            metric_name=self.name,
            value=value,
            group=group,
            step=step,
            timestamp=timestamp,
            label=label,
        )



import typer
import os
import click
import urllib.parse

import socket
import threading
import time
from http.server import ThreadingHTTPServer, SimpleHTTPRequestHandler

from continual.python.sdk.client import Client, __version__

from pathlib import Path


from continual.python.cli import (
    models,
    model_versions,
    experiments,
    projects,
    environments,
    organizations,
    users,
    batch_predictions,
    promotions,
    utils,
    events,
    reports,
    config,
    runs,
)
from typing import Optional
from continual.python.cli.utils import exit_with_error


def version_callback(value: bool):
    if value:
        try:
            c = Client()
            client_version, server_version, upgrade_required = c.version()
            typer.echo(f"Continual SDK Version: {client_version}")
            typer.echo(f"Continual Server Version: {server_version}")
            if upgrade_required:
                typer.secho(
                    f"Upgrade is required",
                    fg="red",
                )
        except Exception as e:
            typer.echo(f"Continual SDK Version: {__version__}")
            typer.secho(
                f"Unable to retrieve Continual Server Version",
                fg="red",
            )
            exit_with_error(e)
        raise typer.Exit()


# ************************#
# Define Typer Structure #
# ************************#


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx):
        return self.commands.keys()


app_helptext = "Continual - Operational AI. Simplified."

cli = typer.Typer(cls=NaturalOrderGroup, help=app_helptext, no_args_is_help=True)

cli.add_typer(models.app, name="models")
cli.add_typer(model_versions.app, name="model-versions")
cli.add_typer(experiments.app, name="experiments")
cli.add_typer(promotions.app, name="promotions")
cli.add_typer(batch_predictions.app, name="batch-predictions")
cli.add_typer(projects.app, name="projects")
cli.add_typer(environments.app, name="environments")
cli.add_typer(organizations.app, name="organizations")
cli.add_typer(users.app, name="users")
cli.add_typer(events.app, name="events")
cli.add_typer(runs.app, name="runs")
cli.add_typer(reports.app, name="report", hidden=True)
cli.add_typer(config.app, name="config")


@cli.callback(context_settings=dict(help_option_names=["-h", "--help"]))
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print Continual CLI version",
    ),
):
    return


@cli.command("help", add_help_option=False, options_metavar="")
def help(ctx: typer.Context):
    """Show CLI usage help."""
    ctx.info_name = None
    typer.echo(ctx.parent.command.get_help(ctx))


def get_verification_token_oob(params: dict) -> str:
    endpoint = utils.get_app_url()
    sso_link = "{}/s/sso/auth?mode=cli-oob".format(endpoint)
    for k, v in params.items():
        sso_link += f"&{k}={urllib.parse.quote_plus(v)}"

    typer.secho("\nVisit the following link to log in:")
    typer.secho(f"\n\t{sso_link}", fg="blue")

    typer.echo("")

    verification_code = ""
    try:
        verification_code = typer.prompt("Verification code", hide_input=True)
    except typer.Abort:
        typer.secho("Aborted!")
        exit(1)

    if not verification_code:
        typer.secho("Invalid verification token.", fg="red")
        exit(1)

    return verification_code


def find_open_port() -> int:
    with socket.socket() as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def is_ssh(echo: bool = False) -> bool:
    val = os.getenv("SSH_CLIENT") is not None or os.getenv("SSH_TTY") is not None
    if echo:
        typer.secho(f"is SSH?: {val}")
    return val


def is_docker(echo: bool = False) -> bool:
    cgroup = Path("/proc/self/cgroup")
    val = (
        Path("/.dockerenv").is_file()
        or cgroup.is_file()
        and cgroup.read_text().find("docker") > -1
    )
    if echo:
        typer.secho(f"is Docker?: {val}")
    return val


def get_verification_token_inband(params: dict) -> str:
    open_port = find_open_port()

    endpoint = utils.get_app_url()
    sso_link = "{}/s/sso/auth?mode=cli&next={}".format(
        endpoint,
        urllib.parse.quote_plus(f"http://localhost:{open_port}/continual-auth"),
    )
    for k, v in params.items():
        sso_link += f"&{k}={urllib.parse.quote_plus(v)}"

    verification_token = ""

    class AuthHandler(SimpleHTTPRequestHandler):

        verification_token = ""

        def log_message(self, format, *args):
            pass

        def do_GET(self):
            if self.path.startswith("/continual-auth"):
                if self.path.split("?")[-1]:
                    query = urllib.parse.parse_qs(self.path.split("?")[-1])
                    if "token" in query:
                        AuthHandler.verification_token = query["token"][0]
                        self.send_response(301)
                        self.send_header(
                            "Location", f"{endpoint}/notifications/cli-login-success"
                        )
                        self.end_headers()

                self.send_response(301)
                self.send_header(
                    "Location", f"{endpoint}/notifications/cli-login-failure"
                )
                self.end_headers()
            else:
                super().do_GET()
            self.server.shutdown()

    class AuthServer(threading.Thread):
        def run(self):
            self.server = ThreadingHTTPServer(("localhost", open_port), AuthHandler)
            self.server.serve_forever()

        def stop(self):
            self.server.shutdown()

    server = AuthServer()
    server.setDaemon(True)
    server.start()

    typer.secho("\nVisit the following link to log in:")
    typer.secho(f"\n\t{sso_link}", fg="blue")

    try:
        typer.secho("\nWaiting for response...")
        while server.is_alive():
            time.sleep(1)
    except typer.Abort:
        typer.echo("Aborted!")
        exit(1)

    verification_token = AuthHandler.verification_token or verification_token
    if not verification_token:
        typer.secho("Invalid verification token.", fg="red")
        exit(1)

    return verification_token


def verify_token(client: Client, token: str) -> str:
    from continual.rpc.management.v1 import management_pb2

    old_api_key = client.config._api_key
    client.config._api_key = ""
    req = management_pb2.ExchangeTokenRequest(token=token)
    res = client._management.ExchangeToken(req)
    if res.token:
        client.config._api_key = old_api_key
        return res.token
    client.config._api_key = old_api_key
    return ""


def sso_login(client: Client, org: str = "", oob: bool = None):
    if oob is None:
        oob = is_ssh() or is_docker()
    try:
        org = typer.prompt("Organization name (organizations/abc)")
    except typer.Abort:
        typer.secho("Aborted!")
        exit(1)

    verification_code = ""
    if oob:
        verification_code = get_verification_token_oob(dict(org=org))
    else:
        verification_code = get_verification_token_inband(dict(org=org))

    api_key = verify_token(client, verification_code)

    old_api_key = client.config._api_key
    client.set_config_api_key(api_key, save=False)

    organization = None

    try:
        user = client.viewer()
        organization = client.organizations.get(org.split("/")[-1])
        typer.secho(
            f"\n🚀 Hey {user.first_name}! We've successfully logged you in to {organization.display_name} (id: {organization.id}) via SAML SSO.\n",
            fg="blue",
        )
        client.config.save()
    except Exception:
        typer.secho("Failed to authenticate user.", fg="red")
        client.config.set_api_key(old_api_key)
        exit(1)

    if organization:
        client.config._project = None
        projects = organization.projects.list_all()
        i = 1
        proj_list = []
        if len(list(projects)) > 0:
            for project in projects:
                if i == 1:
                    typer.echo("Select default project:")
                project_id = project.name.split("/")[-1]
                proj_list.append(project.name)
                typer.echo(f" [{i}]: {project.display_name} ({project_id})")
                i += 1
            key = typer.prompt("Choice")
            client.config._project = proj_list[int(key) - 1]
            client.config._environment = "production"
            client.config.save()
        else:
            typer.secho(
                "This organization has no associated projects. You should create a project in the Continual Web UI and then use `continual config set-project <project>` to use the CLI."
            )
        exit(0)
    else:
        typer.secho("Could not find current organization.", fg="red")
        exit(1)


def oauth_login(
    client: Client, oob: bool = None, google: bool = False, microsoft: bool = False
):
    if oob is None:
        oob = is_ssh() or is_docker()

    provider = ""
    if google:
        provider = "GoogleOAuth"
    elif microsoft:
        provider = "MicrosoftOAuth"
    else:
        typer.secho("An OAuth provider was not specified.", fg="red")

    verification_code = ""
    if oob:
        verification_code = get_verification_token_oob(dict(provider=provider))
    else:
        verification_code = get_verification_token_inband(dict(provider=provider))

    api_key = verify_token(client, verification_code)

    old_api_key = client.config.api_key
    client.set_config_api_key(api_key, save=False)

    try:
        user = client.viewer()
        typer.secho(
            f"\n🚀 Hey {user.first_name}! We've successfully logged you in to Continual via {'Google' if google else 'Microsoft'} OAuth.\n",
            fg="blue",
        )
        client.config.save()
    except Exception:
        typer.secho("Failed to authenticate user.", fg="red")
        client.config.set_api_key(old_api_key)
        exit(1)


@cli.command("login")
@utils.exit_on_error
def login(
    email: str = typer.Option(default="", help="Email address."),
    password: str = typer.Option(default="", help="Password."),
    project: str = typer.Option(None, help="Default project."),
    sso: bool = typer.Option(False, "--sso", help="Log in with SAML SSO."),
    oauth: bool = typer.Option(
        False, "--oauth", help="Log in with a supported OAuth provider."
    ),
    google: bool = typer.Option(False, "--google", help="Log in with Google OAuth."),
    microsoft: bool = typer.Option(
        False, "--microsoft", help="Log in with Microsoft OAuth."
    ),
    endpoint: str = typer.Option(
        None,
        envvar="CONTINUAL_ENDPOINT",
        help="Optional endpoint for private cloud installations.",
    ),
):
    """Log in to Continual.

    Logs in to Continual and saves the session information to
    ~/.continual/continual.yaml which is used by the SDK/CLI. All
    future commands will use the session information to authenticate
    and use the default project if not explicitly passed.
    """
    client = Client(endpoint=endpoint, verify=False)
    if sso:
        return sso_login(client)
    elif google or microsoft or oauth:
        if oauth and not (google or microsoft):
            typer.secho("Pick an OAuth provider to log in with:\n")
            choices = ["Google OAuth", "Microsoft OAuth"]
            for idx, choice in enumerate(choices):
                typer.echo(f" [{idx + 1}]: {choice}")
            typer.echo("")
            try:
                key = typer.prompt("Choice")
                oauth_choice = choices[int(key) - 1]
                if oauth_choice == "Google OAuth":
                    google = True
                elif oauth_choice == "Microsoft OAuth":
                    microsoft = True
            except typer.Abort:
                typer.secho("Aborted!")
                exit(1)
        return oauth_login(client, google=google, microsoft=microsoft)
    else:
        if not email:
            try:
                email = typer.prompt("Email")
            except typer.Abort:
                typer.secho("Aborted!")
                exit(1)

        if not password:
            try:
                password = typer.prompt("Password", hide_input=True)
            except typer.Abort:
                typer.secho("Aborted!")
                exit(1)

    user = client.login(email=email, password=password)
    if project is None:
        client.config._project = None
        projects = client.projects.list_all()
        i = 1
        proj_list = []
        if len(list(projects)) > 0:
            for project in projects:
                if i == 1:
                    typer.echo("Select default project:")
                project_id = project.name.split("/")[-1]
                proj_list.append(project.name)
                typer.echo(f" [{i}]: {project.display_name} ({project_id})")
                i += 1
            key = typer.prompt("Choice")
            client.config._project = proj_list[int(key) - 1]
            client.config._environment = "production"
        else:
            typer.secho(
                "You have no projects created in Continual. You should create a project in the Continual Web UI and then use `continual config set-project <project>` to use the CLI."
            )
    else:
        client.config._project = project
        client.config._environment = "production"
    client.config.save()
    typer.echo(
        f"🚀 Login successful.  Welcome to Continual, {user.first_name} {user.last_name}."
    )


@cli.command("logout")
@utils.exit_on_error
def logout():
    """Log out of Continual.

    Logs out of the current session.
    """
    c = Client(verify=False)
    c.logout()
    typer.echo("👋 See you later, come back soon!")


if __name__ == "__main__":
    cli()



import typer

from continual.python.cli import utils
from continual.python.sdk.client import Client


app = typer.Typer(help="Manage CLI configuration.")

# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        show()


@app.command("clear-all")
@utils.exit_on_error
def continual_clear_all(
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Clear all configuration options.

    Clear all configurations and save the new configuration.
    """

    try:
        continual_clear(
            project=True,
            environment=True,
            style=True,
            endpoint=True,
            api_key=True,
            print=print,
        )
    except:
        typer.secho(
            "Error clearing all of the config",
            fg="red",
        )


@app.command("clear")
@utils.exit_on_error
def continual_clear(
    project: bool = typer.Option(False, "--project", help="Project ID."),
    environment: bool = typer.Option(False, "--environment", help="Environment Id"),
    style: bool = typer.Option(False, "--style", help="Display style."),
    endpoint: bool = typer.Option(
        False, "--endpoint", help="Optional endpoint for private cloud installations."
    ),
    api_key: bool = typer.Option(False, "--api-key", help="Api Key."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Clear configurations.

    Clear specific configurations and save the new configuration.
    """

    try:
        client = Client(verify=False)
        if project:
            client.config._project = ""
        if environment:
            client.config._environment = ""
        if style:
            client.config._style = "GREEN"
        if endpoint:
            client.config._endpoint = ""
        if api_key:
            client.config._api_key = ""

        client.config.save()

        if print:
            client.config.show()
    except:
        typer.secho(
            "Error clearing config",
            fg="red",
        )


@app.command("init")
@utils.exit_on_error
def continual_init(
    project: str = typer.Option(..., help="Project ID."),
    environment: str = typer.Option(None, help="Environment Id"),
    style: utils.ContinualStyle = typer.Option(None, help="Display style."),
):
    """Initialize configuration.

    Sets the default project, environment, and style for the SDK and CLI.  All future commands
    will use the default project if no project is explicitly passed. You may
    set the project and environment at the same time via project@environment.
    """
    try:
        c = Client(verify=False)
        set_project(project=project, print=False)
        if environment:
            set_environment(environment=environment, print=False)
        if style:
            set_style(style=style, print=False)
        c.config.show()
    except Exception as e:
        typer.secho(
            "Configuration initialization failed: %s" % e,
            fg="red",
        )


@app.command("set-project")
@utils.exit_on_error
def set_project(
    project: str = typer.Argument(..., help="Project ID."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default project.

    Sets the default project for the SDK and CLI.  All future commands
    will use the default project if no project is explicitly passed. You may
    set the project and environment at the same time via project@environment.
    """
    try:
        client = Client(verify=False)
        client.set_config_project(project, save=True)
        if print:
            client.config.show()
    except:
        typer.secho(
            "Project %s does not exist. Please create the project first!" % project,
            fg="red",
        )


@app.command("set-endpoint")
@utils.exit_on_error
def set_endpoint(
    endpoint: str = typer.Argument(..., help="Continual environment endpoint"),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default endpoint"""
    c = Client(verify=False)
    old_endpoint = c.config.endpoint
    c.config._endpoint = endpoint
    c.config.save()

    # Reload client to check endpoint validity
    try:
        c = Client(verify=False)
        typer.secho(f"Sucessfully set endpoint to {endpoint}", fg="green")
        if print:
            c.config.show()
    except ValueError as e:
        c.config._endpoint = old_endpoint
        c.config.save()
        typer.secho("Endpoint %s is not valid : %s" % (endpoint, e), fg="red")


@app.command("set-environment")
@utils.exit_on_error
def set_environment(
    environment: str = typer.Argument(..., help="Environment ID."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default environment.

    Sets the default environment for the SDK and CLI.  All future commands
    will use the default environment if no environment is explicitly passed.
    """
    try:
        client = Client(verify=False)
        client.set_config_environment(environment=environment, save=True)
        if print:
            client.config.show()
    except:
        typer.secho(
            "Environment %s does not exist. Please create the environment first!"
            % environment,
            fg="red",
        )


@app.command("set-style")
@utils.exit_on_error
def set_style(
    style: utils.ContinualStyle = typer.Argument(..., help="Display style."),
    print: bool = typer.Option(True, help="Print config after modification."),
):
    """Set default CLI display style."""
    c = Client(verify=False)
    c.config._style = style.value
    c.config.save()
    if print:
        c.config.show()


@app.command("show")
@utils.exit_on_error
def show():
    """Show current config.

    Shows the current session configuration
    """
    c = Client(verify=False)
    c.config.show()



from __future__ import annotations
from typing import List, Optional
import json
from continual.python.sdk.iterators import Pager

from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.events import EventManager
from continual.python.sdk.metadata import NpEncoder
from continual.rpc.management.v1 import management_pb2, types

CHUNK_SIZE = 117440512  # 112 MB in bytes, chosen arbitrarily


class ResourceChecksManager(Manager):
    """Manages resource check resources."""

    # the name pattern for a resource check depends on the resource it was created for
    name_pattern: str = ""

    def get(self, id: str) -> ResourceCheck:
        """Get resource check.
        Arguments:
            id: ResourceCheck name or id.
        Returns
            A ResourceCheck.
        Examples:
            >>> ... # Assume client, project, environment, and dataset are defined
            >>> dataset_version = dataset.dataset_versions.create()
            >>> dc = dataset_version.resource_checks.create(display_name="Check Null Values", group_name="test", outcome="PASSED", summary="No null values", state="PASS")
            >>> dataset_version.resource_checks.get(id=dc.id)
            <ResourceCheck object {'name': 'projects/continual_test_proj/environments/production/datasets/test_dataset/versions/cegaeq25lsrt9r5a8ma0/resourceChecks/cegai425lsrt9r5a8md0',
            'run_name': 'projects/continual_test_proj/environments/production/runs/cegaeq25lsrt9r5a8m80', 'group_name': 'test', 'outcome': 'PASSED',
            'summary': 'No null values', 'create_time': '2022-12-19T18:10:24.774076Z', 'display_name': 'Check Null Values', 'duration': 0.0,
            'errors': [], 'warnings': [], 'infos': [], 'artifact_name': ''}>
        """

        req = management_pb2.GetResourceCheckRequest(name=self.name(id))
        resource_check = self.client._management.GetResourceCheck(req)
        return ResourceCheck.from_proto(resource_check, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
    ) -> List[ResourceCheck]:
        """List resource checks.
        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
        Returns:
            A list of ResourceChecks.
        Examples:
            >>> ... # Assume client, project, environment, and dataset are defined
            >>> dataset_version = dataset.dataset_versions.create()
            >>> checks = [dataset_version.resource_checks.create(display_name=f"Test {i}", group_name="test", outcome=f"{('PASSED' if i%2 == 1 else 'FAILED')}", summary=f"Test {i} passed.") for i in range(5)]
            >>> [check.outcome for check in dataset_version.resource_checks.list(page_size=10)]
            [<CheckOutcome.FAILED: 'FAILED'>, <CheckOutcome.PASSED: 'PASSED'>, <CheckOutcome.FAILED: 'FAILED'>, <CheckOutcome.PASSED: 'PASSED'>, <CheckOutcome.FAILED: 'FAILED'>]
            >>> [check.outcome for check in dataset_version.resource_checks.list(page_size=10, order_by='outcome')]
            [<CheckOutcome.FAILED: 'FAILED'>, <CheckOutcome.PASSED: 'PASSED'>]
        """
        req = management_pb2.ListResourceChecksRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListResourceChecks(req)
        return [
            ResourceCheck.from_proto(x, client=self.client)
            for x in resp.resource_checks
        ]

    def list_all(self) -> Pager[ResourceCheck]:
        """List all resource checks.
        Pages through all resource_checks using an iterator.
        Returns:
            A iterator of all resource_checks.
        Examples:
            >>> ... # Assume client, project, environment, and dataset are defined
            >>> dataset_version = dataset.dataset_versions.create()
            >>> checks = [dataset_version.resource_checks.create(display_name=f"Test {i}", group_name="test", outcome='PASSED', summary=f"Test {i} passed.") for i in range(5)]
            >>> [check.summary for check in dataset_version.resource_checks.list_all()]
            ['Test 0 passed.', 'Test 1 passed.', 'Test 2 passed.', 'Test 3 passed.', 'Test 4 passed.']
        """

        def next_page(next_page_token):
            req = management_pb2.ListResourceChecksRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListResourceChecks(req)
            return (
                [
                    ResourceCheck.from_proto(x, client=self.client)
                    for x in resp.resource_checks
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def create(
        self,
        display_name: str,
        outcome: str,
        group_name: str = "",
        duration: float = 0.0,
        data: dict = {},
        errors: List[str] = [],
        warnings: List[str] = [],
        infos: List[str] = [],
        summary: str = "",
        artifact_name: Optional[str] = "",
    ) -> ResourceCheck:
        """Create a ResourceCheck.
        Arguments:
            group_name: Name of the group.
            outcome: whether the check PASSED, FAILED, or was SKIPPED.
            duration: Duration of the resource check.
            data: Dictionary of additional information if desired.
            errors: List of errors.
            warnings: List of warnings.
            infos: List of infos.
            summary: Human readable summary of the resource check result.
            artifact_name: Name of an artifact associated with the resource check.
            display_name: Display name of the resource check.
        Returns:
            ResourceCheck object.
        Examples:
            >>> ... # Assume client, project, environment, and dataset are defined
            >>> dataset_version = dataset.dataset_versions.create()
            >>> dataset_version.resource_checks.create(display_name="Check Null Values", group_name="test", outcome="PASSED"", summary="No null values", state="PASS")
            <ResourceCheck object {'name': 'projects/continual_test_proj/environments/production/datasets/test_dataset/versions/cegaeq25lsrt9r5a8ma0/resourceChecks/cegai425lsrt9r5a8md0',
            'run_name': 'projects/continual_test_proj/environments/production/runs/cegaeq25lsrt9r5a8m80', 'group_name': 'test', 'outcome': 'PASSED',
            'summary': 'No null values', 'create_time': '2022-12-19T18:10:24.774076Z', 'display_name': 'Check Null Values', 'duration': 0.0, 'data': {},
            'errors': [], 'warnings': [], 'infos': [], 'artifact_name': ''}>
        """
        req = management_pb2.CreateResourceCheckRequest(
            parent=self.parent,
            resource_check=ResourceCheck(
                display_name=display_name,
                group_name=group_name,
                outcome=outcome,
                duration=duration,
                data=json.dumps(data, cls=NpEncoder),
                errors=errors,
                warnings=warnings,
                infos=infos,
                summary=summary,
                artifact_name=artifact_name,
                run_name=self.run_name,
            ).to_proto(),
        )
        resp = self.client._management.CreateResourceCheck(req)
        return ResourceCheck.from_proto(resp, client=self.client)


class ResourceCheck(Resource, types.ResourceCheck):
    """ResourceCheck resource."""

    # the name pattern for resource check depends on the resource it was created for
    name_pattern: str = ""
    manager: ResourceChecksManager

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = ResourceChecksManager(parent=self.parent, client=self.client)



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.projects import ProjectManager
from google.protobuf import field_mask_pb2


class OrganizationManager(Manager):
    """Manages user resources."""

    name_pattern: str = "organizations/{user}"

    def create(self, display_name: str) -> Organization:
        """Create an organization.

        Arguments:
            display_name: Display name.

        Returns:
            A new organization.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> client.organizations.create(display_name="example-org")
            <Organization object {'name': 'organizations/ced8rci5lsrrfj0t5f30', 'display_name': 'example-org',
            'update_time': '2022-12-15T03:00:02.419421Z', 'create_time': '2022-12-15T03:00:02.419421Z',
            'status': 'EXPIRED', 'trial_credits': '0', 'trial_credits_used': '0', 'show_plan': False,
            'sso_enabled': False, 'requires_sso': False, 'sso_domains': [], 'allow_external_users': False,
            'sso_configured': False, 'directory_sync_configured': False}>
        """
        req = management_pb2.CreateOrganizationRequest(
            organization=types.Organization(display_name=display_name).to_proto()
        )
        resp = self.client._management.CreateOrganization(req)
        return Organization.from_proto(resp, client=self.client)

    def get(self, id: str) -> Organization:
        """Get organization.

        Arguments:
            id: Organization name or id.

        Returns
            An organization.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> client.organizations.get(id=org.name)
            <Organization object {'name': 'organizations/ced8rci5lsrrfj0t5f30', 'display_name': 'example-org',
            'update_time': '2022-12-15T03:00:02.419421Z', 'create_time': '2022-12-15T03:00:02.419421Z',
            'status': 'EXPIRED', 'trial_credits': '0', 'trial_credits_used': '0', 'show_plan': False,
            'sso_enabled': False, 'requires_sso': False, 'sso_domains': [], 'allow_external_users': False,
            'sso_configured': False, 'directory_sync_configured': False}>
        """
        req = management_pb2.GetOrganizationRequest(name=self.name(id))
        resp = self.client._management.GetOrganization(req)
        return Organization.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
    ) -> List[Organization]:
        """List organizations.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of Organizations.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> orgs = [client.organizations.create(display_name=f"example-org-{i}") for i in range(10)]
            >>> [org.display_name for org in client.organizations.list(page_size=10)] # Orders by 'create_time' by default
            ['example-org-0', 'example-org-1', 'example-org-2', 'example-org-3', 'example-org-4', 'example-org-5',
            'example-org-6', 'example-org-7', 'example-org-8', 'example-org-9']
            >>> [org.display_name for org in client.organizations.list(page_size=10, latest=False)] # Ascending order of create_time
            ['test2@continual.ai', 'test@continual.ai']
        """
        req = management_pb2.ListOrganizationsRequest(
            page_size=page_size, order_by=order_by, latest=latest
        )
        resp = self.client._management.ListOrganizations(req)
        return [
            Organization.from_proto(u, client=self.client) for u in resp.organizations
        ]

    def list_all(self) -> Pager[Organization]:
        """List all organizations.

        Pages through all organization using an iterator.

        Returns:
            A iterator of all organizations.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> orgs = [client.organizations.create(display_name=f"example-org-{i}") for i in range(10)]
            >>> [org.display_name for org in client.organizations.list_all()] # Orders by 'create_time' by default
            ['example-org-0', 'example-org-1', 'example-org-2', 'example-org-3', 'example-org-4', 'example-org-5',
            'example-org-6', 'example-org-7', 'example-org-8', 'example-org-9']
        """

        def next_page(next_page_token):
            req = management_pb2.ListOrganizationsRequest(page_token=next_page_token)
            resp = self.client._management.ListOrganizations(req)
            return (
                [
                    Organization.from_proto(u, client=self.client)
                    for u in resp.organizations
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(self, id: str) -> None:
        """Delete an organization.

        Arguments:
            id: Organization name or id.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> orgs = [client.organizations.create(display_name=f"example-org-{i}") for i in range(10)]
            >>> [client.organizations.delete(id=org.id)  for org in client.organizations.list_all()]
            [None, None, None, None, None, None, None, None, None, None]
            >>> len([org  for org in client.organizations.list_all()])
            0
        """
        req = management_pb2.DeleteOrganizationRequest(name=self.name(id))
        self.client._management.DeleteOrganization(req)

    def update(
        self,
        id: str,
        display_name: Optional[str] = None,
    ) -> Organization:
        """Update organization.

        Arguments:
            id: The organization's id
            display_name:  Display name.

        Returns:
            Updated organization.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="name_wiht_typo")
            >>> client.organizations.update(id=org.name, display_name="name_without_typo").display_name
            'name_without_typo'
        """
        paths = []
        if display_name is not None:
            paths.append("display_name")
        req = management_pb2.UpdateOrganizationRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            organization=Organization(
                name=self.name(id),
                display_name=display_name,
            ).to_proto(),
        )
        resp = self.client._management.UpdateOrganization(req)
        return Organization.from_proto(resp, client=self.client)


class Organization(Resource, types.Organization):
    """Organization resource."""

    name_pattern = "organizations/{users}"

    _manager: OrganizationManager

    _projects: ProjectManager

    def _init(self):
        self._manager = OrganizationManager(parent=self.parent, client=self.client)
        self._projects = ProjectManager(parent=self.name, client=self.client)

    @property
    def projects(self) -> ProjectManager:
        """Organization's Project Manager."""
        return self._projects

    def delete(self) -> None:
        """Delete organization.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> org.delete()
            >>> len([org for org in client.organizations.list_all()])
            0
        """
        self._manager.delete(self.name)

    def update(
        self,
        display_name: Optional[str] = None,
    ) -> Organization:
        """Update organization.

        Arguments:
            display_name:  Display name.

        Returns:
            Updated organization.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="name_wiht_typo")
            >>> org = org.update(display_name="name_without_typo")
            >>> org.display_name
            'name_without_typo'
        """
        return self._manager.update(
            self.name,
            display_name=display_name,
        )

    def _create_user_role(self, user_name: str, role: str) -> None:
        """Create an organization role for a user

        Arguments:
            user_name: the name of the user
            role: the name of the role
        """

        req = management_pb2.CreateAccessPolicyRequest(
            parent=self.name,
            access_policy=types.AccessPolicy(
                resource=self.name, subject=user_name, role=role
            ).to_proto(),
        )
        self._manager.client._management.CreateAccessPolicy(req)
